"""Generation helpers for DependencyGraphGenerator."""

from __future__ import annotations

from typing import TYPE_CHECKING

from yaraast.metrics.dependency_graph_graphviz import (
    add_complexity_legend,
    add_complexity_nodes,
    add_module_edges,
    add_module_nodes,
    add_module_rule_nodes,
    add_rule_graph_nodes,
    add_rule_string_edges,
    create_graph,
)
from yaraast.metrics.dependency_graph_helpers import render_graph, reset_graph_state
from yaraast.metrics.dependency_graph_render import (
    complexity_node_color,
    complexity_node_label,
    rule_graph_label,
    rule_node_color,
)

if TYPE_CHECKING:
    from yaraast.ast.base import YaraFile
    from yaraast.metrics.dependency_graph import DependencyGraphGenerator


def generate_graph(
    generator: DependencyGraphGenerator,
    ast: YaraFile,
    output_path: str | None = None,
    format: str = "svg",
    engine: str = "dot",
) -> str:
    reset_graph_state(generator)
    generator.visit(ast)
    dot = create_graph("YARA Dependencies", rankdir="TB", engine=engine)
    generator._add_nodes(dot)
    generator._add_edges(dot)
    return render_graph(dot, output_path, format)


def generate_rule_graph(
    generator: DependencyGraphGenerator,
    ast: YaraFile,
    output_path: str | None = None,
    format: str = "svg",
) -> str:
    generator.visit(ast)
    dot = create_graph("YARA Rule Dependencies", rankdir="LR")
    from yaraast.metrics.dependency_graph_graphviz import apply_rule_graph_style

    apply_rule_graph_style(dot)
    add_rule_graph_nodes(dot, generator.rules, rule_graph_label, rule_node_color)
    add_rule_string_edges(dot, generator.string_references)
    return render_graph(dot, output_path, format)


def generate_module_graph(
    generator: DependencyGraphGenerator,
    ast: YaraFile,
    output_path: str | None = None,
    format: str = "svg",
) -> str:
    generator.visit(ast)
    dot = create_graph("YARA Module Dependencies", rankdir="TB")
    add_module_nodes(dot, generator.imports)
    add_module_rule_nodes(dot, generator.rules)
    add_module_edges(dot, generator.module_references, generator.imports)
    return render_graph(dot, output_path, format)


def generate_complexity_graph(
    generator: DependencyGraphGenerator,
    ast: YaraFile,
    complexity_metrics: dict[str, int],
    output_path: str | None = None,
    format: str = "svg",
) -> str:
    generator.visit(ast)
    dot = create_graph("YARA Complexity Visualization", rankdir="TB")
    add_complexity_nodes(
        dot,
        generator.rules,
        complexity_metrics,
        complexity_node_label,
        complexity_node_color,
    )
    add_complexity_legend(dot)
    return render_graph(dot, output_path, format)



"""Real CLI tests for performance commands (no mocks)."""

from __future__ import annotations

import json
import os
import signal
import stat
import sys
import threading
from pathlib import Path
from textwrap import dedent

import pytest
from click.testing import CliRunner

from yaraast.cli.commands.performance import performance


def _write(tmp_path, name: str, content: str) -> str:
    path = tmp_path / name
    path.write_text(dedent(content))
    return str(path)


def _sample_yara() -> str:
    return """
    rule perf_rule {
        strings:
            $a = "abc"
        condition:
            $a
    }
    """


def test_performance_batch_file(tmp_path) -> None:
    file_path = _write(tmp_path, "rule.yar", _sample_yara())
    out_dir = tmp_path / "batch_out"
    runner = CliRunner()

    result = runner.invoke(
        performance,
        [
            "batch",
            file_path,
            "--output-dir",
            str(out_dir),
            "--operations",
            "parse",
        ],
    )

    assert result.exit_code == 0
    assert out_dir.exists()
    results_file = out_dir / "batch_results.json"
    assert results_file.exists()
    payload = json.loads(results_file.read_text())
    assert "parse" in payload


def test_performance_stream_file(tmp_path) -> None:
    file_path = _write(tmp_path, "rule.yar", _sample_yara())
    out_file = tmp_path / "stream_results.json"
    runner = CliRunner()

    result = runner.invoke(
        performance,
        [
            "stream",
            file_path,
            "--output",
            str(out_file),
        ],
    )

    assert result.exit_code == 0
    assert out_file.exists()
    payload = json.loads(out_file.read_text())
    assert payload["summary"]["total_processed"] >= 1


def test_performance_parallel_and_optimize(tmp_path) -> None:
    file_path = _write(tmp_path, "rule.yar", _sample_yara())
    out_dir = tmp_path / "parallel_out"
    runner = CliRunner()

    parallel = runner.invoke(
        performance,
        [
            "parallel",
            file_path,
            "--output-dir",
            str(out_dir),
            "--analysis-type",
            "complexity",
            "--chunk-size",
            "1",
        ],
    )

    assert parallel.exit_code == 0
    assert out_dir.exists()
    assert (out_dir / "complexity_analysis.json").exists()

    optimize = runner.invoke(
        performance,
        ["optimize", "10", "--memory-mb", "256", "--target-time", "5"],
    )

    assert optimize.exit_code == 0
    assert "Optimization Recommendations" in optimize.output


def test_performance_batch_uses_default_output_dir_and_progress(tmp_path) -> None:
    first = _write(tmp_path, "one.yar", _sample_yara())
    _write(tmp_path, "two.yar", _sample_yara().replace("perf_rule", "perf_rule_two"))
    runner = CliRunner()

    result = runner.invoke(
        performance,
        [
            "batch",
            str(tmp_path),
            "--operations",
            "parse",
            "--recursive",
            "--progress",
        ],
    )

    assert result.exit_code == 0
    assert "%" in result.output
    assert (tmp_path.parent / f"{tmp_path.name}_batch_output" / "batch_results.json").exists()
    assert first


def test_performance_batch_directory_without_progress_covers_silent_callback(tmp_path) -> None:
    _write(tmp_path, "one.yar", _sample_yara())
    _write(tmp_path, "two.yar", _sample_yara().replace("perf_rule", "silent_rule"))
    runner = CliRunner()
    out_dir = tmp_path / "silent_out"

    result = runner.invoke(
        performance,
        [
            "batch",
            str(tmp_path),
            "--output-dir",
            str(out_dir),
            "--operations",
            "parse",
            "--recursive",
        ],
    )

    assert result.exit_code == 0
    assert (out_dir / "batch_results.json").exists()


def test_performance_stream_progress_without_output_file(tmp_path) -> None:
    file_path = _write(tmp_path, "rule.yar", _sample_yara())
    runner = CliRunner()

    result = runner.invoke(
        performance,
        [
            "stream",
            file_path,
            "--progress",
        ],
    )

    assert result.exit_code == 0
    assert "%" in result.output
    assert "rule.yar" in result.output


def test_performance_parallel_empty_directory_and_failed_parse(tmp_path) -> None:
    runner = CliRunner()
    empty_dir = tmp_path / "empty"
    empty_dir.mkdir()

    result = runner.invoke(performance, ["parallel", str(empty_dir)])
    assert result.exit_code == 0
    assert "No YARA files found to process" in result.output

    bad_dir = tmp_path / "bad"
    bad_dir.mkdir()
    (bad_dir / "broken.yar").write_text("rule broken { condition: ", encoding="utf-8")
    result = runner.invoke(
        performance, ["parallel", str(bad_dir), "--output-dir", str(tmp_path / "bad_out")]
    )
    assert result.exit_code == 0
    assert "No files parsed successfully" in result.output


def test_performance_parallel_dependency_and_default_output_dir(tmp_path) -> None:
    runner = CliRunner()

    with runner.isolated_filesystem(temp_dir=str(tmp_path)):
        rule_path = Path("rule.yar")
        rule_path.write_text(dedent(_sample_yara()).strip() + "\n", encoding="utf-8")

        result = runner.invoke(
            performance,
            [
                "parallel",
                str(rule_path),
                "--analysis-type",
                "dependency",
                "--chunk-size",
                "1",
            ],
        )

        assert result.exit_code == 0
        assert "Generated 1 dependency graphs" in result.output
        assert Path("parallel_analysis_output").exists()


@pytest.mark.skipif(sys.platform == "win32", reason="chmod read-only not effective on Windows")
def test_performance_batch_and_stream_abort_on_real_write_errors(tmp_path) -> None:
    runner = CliRunner()
    file_path = _write(tmp_path, "rule.yar", _sample_yara())

    ro_dir = tmp_path / "readonly_batch"
    ro_dir.mkdir()
    ro_dir.chmod(stat.S_IREAD | stat.S_IEXEC)
    try:
        batch = runner.invoke(
            performance,
            [
                "batch",
                file_path,
                "--output-dir",
                str(ro_dir),
                "--operations",
                "parse",
            ],
        )
    finally:
        ro_dir.chmod(stat.S_IRWXU)

    assert batch.exit_code == 1
    assert "Error during batch processing" in batch.output

    stream = runner.invoke(
        performance,
        [
            "stream",
            file_path,
            "--output",
            str(tmp_path / "missing" / "stream.json"),
        ],
    )
    assert stream.exit_code == 1
    assert "Error during streaming parse" in stream.output


@pytest.mark.skipif(sys.platform == "win32", reason="chmod read-only not effective on Windows")
def test_performance_parallel_abort_on_real_complexity_write_error(tmp_path) -> None:
    runner = CliRunner()
    file_path = _write(tmp_path, "rule.yar", _sample_yara())
    ro_dir = tmp_path / "readonly_parallel"
    ro_dir.mkdir()
    ro_dir.chmod(stat.S_IREAD | stat.S_IEXEC)
    try:
        result = runner.invoke(
            performance,
            [
                "parallel",
                file_path,
                "--output-dir",
                str(ro_dir),
                "--analysis-type",
                "complexity",
                "--chunk-size",
                "1",
            ],
        )
    finally:
        ro_dir.chmod(stat.S_IRWXU)

    assert result.exit_code == 1
    assert "Error during parallel analysis" in result.output


@pytest.mark.skipif(sys.platform == "win32", reason="SIGINT via os.kill not reliable on Windows")
def test_performance_stream_and_parallel_handle_real_sigint(tmp_path) -> None:
    rules_dir = tmp_path / "sigint_rules"
    rules_dir.mkdir()
    for i in range(2000):
        (rules_dir / f"{i}.yar").write_text(dedent(_sample_yara()).strip() + "\n", encoding="utf-8")

    timer = threading.Timer(0.3, lambda: os.kill(os.getpid(), signal.SIGINT))
    timer.start()
    try:
        stream = CliRunner().invoke(
            performance,
            ["stream", str(rules_dir), "--recursive", "--progress"],
        )
    finally:
        timer.cancel()
    assert stream.exit_code == 0
    assert "Parsing cancelled by user" in stream.output

    timer = threading.Timer(0.5, lambda: os.kill(os.getpid(), signal.SIGINT))
    timer.start()
    try:
        parallel = CliRunner().invoke(
            performance,
            ["parallel", str(rules_dir), "--analysis-type", "complexity", "--chunk-size", "1"],
        )
    finally:
        timer.cancel()
    assert parallel.exit_code == 0
    assert "Analysis cancelled by user" in parallel.output



"""Display helpers for serialize CLI."""

from __future__ import annotations

from typing import Any

from rich.console import Console
from rich.panel import Panel
from rich.table import Table

from yaraast.ast.base import YaraFile
from yaraast.serialization.ast_diff import DiffType

console = Console()


def _display_protobuf_stats(stats: dict[str, Any]) -> None:
    """Display Protobuf serialization statistics."""
    table = Table(title="Protobuf Serialization Stats")
    table.add_column("Metric", style="cyan")
    table.add_column("Value", style="green")

    table.add_row("Binary Size", f"{stats['binary_size_bytes']:,} bytes")
    table.add_row("Text Size", f"{stats['text_size_bytes']:,} bytes")
    table.add_row("Compression Ratio", f"{stats['compression_ratio']:.2f}x")
    table.add_row("Rules Count", str(stats["rules_count"]))
    table.add_row("Imports Count", str(stats["imports_count"]))

    console.print(table)


def _display_diff_summary(diff_result: Any) -> None:
    """Display difference summary table."""
    table = Table(title="AST Differences Summary")
    table.add_column("Change Type", style="cyan")
    table.add_column("Count", style="green", justify="right")

    summary = diff_result.change_summary
    for change_type, count in summary.items():
        if count > 0:
            icon = {
                "added": "+",
                "removed": "-",
                "modified": "📝",
                "moved": "↔️",
                "unchanged": "✅",
            }.get(change_type, "•")
            table.add_row(f"{icon} {change_type.title()}", str(count))

    console.print(table)


def _display_detailed_changes(diff_result: Any) -> None:
    """Display detailed changes if not too many."""
    if len(diff_result.differences) <= 20:
        console.print("\n[bold]Detailed Changes:[/bold]")
        for diff_node in diff_result.differences:
            icon = {
                DiffType.ADDED: "[green]+[/green]",
                DiffType.REMOVED: "[red]-[/red]",
                DiffType.MODIFIED: "[yellow]📝[/yellow]",
                DiffType.MOVED: "[blue]↔️[/blue]",
            }.get(diff_node.diff_type, "•")

            console.print(f"  {icon} {diff_node.path} ({diff_node.node_type})")
            if diff_node.diff_type == DiffType.MODIFIED:
                console.print(f"    [dim]Old:[/dim] {diff_node.old_value}")
                console.print(f"    [dim]New:[/dim] {diff_node.new_value}")
    else:
        console.print(
            f"\n[dim]Use --output to save detailed changes ({len(diff_result.differences)} total)[/dim]",
        )


def _display_diff_statistics(diff_result: Any) -> None:
    """Display comparison statistics."""
    stats_table = Table(title="Comparison Statistics")
    stats_table.add_column("Metric", style="cyan")
    stats_table.add_column("Old", style="red", justify="right")
    stats_table.add_column("New", style="green", justify="right")

    stats_table.add_row(
        "Rules",
        str(diff_result.statistics["old_rules_count"]),
        str(diff_result.statistics["new_rules_count"]),
    )
    stats_table.add_row(
        "Imports",
        str(diff_result.statistics["old_imports_count"]),
        str(diff_result.statistics["new_imports_count"]),
    )
    stats_table.add_row("AST Hash", diff_result.old_ast_hash, diff_result.new_ast_hash)

    console.print(stats_table)


def build_validation_panel(
    input_file: str,
    format: str,
    ast: YaraFile | None,
    error: Exception | None,
) -> Panel:
    if error is None and ast is not None:
        return Panel(
            f"[green]✅ Valid {format.upper()} serialization[/green]\n\n"
            f"📊 Structure:\n"
            f"  • Rules: {len(ast.rules)}\n"
            f"  • Imports: {len(ast.imports)}\n"
            f"  • Includes: {len(ast.includes)}",
            title=f"Validation Result: {input_file}",
            border_style="green",
        )

    error_text = str(error) if error else "Unknown error"
    return Panel(
        f"[red]❌ Invalid {format.upper()} serialization[/red]\n\nError: {error_text}",
        title=f"Validation Result: {input_file}",
        border_style="red",
    )



"""Statistics helpers for string pattern diagrams."""

from __future__ import annotations

from collections import Counter
from typing import Any


class StringDiagramStatsMixin:
    """Mixin providing string diagram helpers."""

    def get_pattern_statistics(self) -> dict[str, Any]:
        """Get comprehensive pattern statistics."""
        if not self.string_patterns:
            return {"total_patterns": 0}

        return {
            "total_patterns": len(self.string_patterns),
            "by_type": {
                "plain": len(
                    [p for p in self.string_patterns.values() if p["type"] == "plain"],
                ),
                "hex": len(
                    [p for p in self.string_patterns.values() if p["type"] == "hex"],
                ),
                "regex": len(
                    [p for p in self.string_patterns.values() if p["type"] == "regex"],
                ),
            },
            "complexity_distribution": self._get_complexity_distribution(),
            "common_patterns": self._find_common_patterns(),
            "pattern_lengths": self._get_length_statistics(),
            "modifiers_usage": self._get_modifier_statistics(),
        }

    def _get_complexity_distribution(self) -> dict[str, int]:
        """Get distribution of complexity scores."""
        distribution = {"low": 0, "medium": 0, "high": 0}

        for pattern_info in self.string_patterns.values():
            complexity = self._calculate_pattern_complexity(pattern_info)

            if complexity <= 3:
                distribution["low"] += 1
            elif complexity <= 6:
                distribution["medium"] += 1
            else:
                distribution["high"] += 1

        return distribution

    def _find_common_patterns(self) -> list[tuple[str, int]]:
        """Find most common pattern types/characteristics."""
        common = []

        # Most common modifiers
        common.extend(
            [("modifier_" + mod, count) for mod, count in self._modifier_counts().most_common(3)],
        )

        return common

    def _get_length_statistics(self) -> dict[str, float]:
        """Get pattern length statistics."""
        lengths = [pattern_info.get("length", 0) for pattern_info in self.string_patterns.values()]

        if not lengths:
            return {}

        return {
            "min": min(lengths),
            "max": max(lengths),
            "avg": sum(lengths) / len(lengths),
            "median": sorted(lengths)[len(lengths) // 2],
        }

    def _get_modifier_statistics(self) -> dict[str, int]:
        """Get modifier usage statistics."""
        return dict(self._modifier_counts())

    def _modifier_counts(self) -> Counter:
        modifier_counts = Counter()
        for pattern_info in self.string_patterns.values():
            for modifier in pattern_info.get("modifiers", []):
                modifier_counts[modifier] += 1
        return modifier_counts



"""UTF-16 position utilities for LSP protocol compliance."""

from __future__ import annotations


def utf16_len(text: str) -> int:
    """Count UTF-16 code units needed to encode text."""
    count = 0
    for ch in text:
        code_point = ord(ch)
        count += 2 if code_point > 0xFFFF else 1
    return count


def utf8_col_to_utf16(line: str, col: int) -> int:
    """Convert a 0-based character column to UTF-16 code unit offset."""
    prefix = line[:col]
    return utf16_len(prefix)


def utf16_col_to_utf8(line: str, utf16_col: int) -> int:
    """Convert a UTF-16 code unit offset to 0-based character column."""
    units = 0
    for i, ch in enumerate(line):
        if units >= utf16_col:
            return i
        code_point = ord(ch)
        units += 2 if code_point > 0xFFFF else 1
    return len(line)



"""Helpers for YAML serializer output shaping."""

from __future__ import annotations

from pathlib import Path
from typing import Any

import yaml

from yaraast.config import YAML_DEFAULT_WIDTH
from yaraast.serialization.serializer_helpers import write_text


def dump_yaml(data: Any, *, flow_style: bool, width: int = YAML_DEFAULT_WIDTH) -> str:
    """Render YAML with consistent formatting defaults."""
    return yaml.dump(
        data,
        default_flow_style=flow_style,
        allow_unicode=True,
        sort_keys=False,
        indent=2,
        width=width,
    )


def serialize_yaml(
    data: Any, output_path: str | Path | None, *, flow_style: bool, width: int = YAML_DEFAULT_WIDTH
) -> str:
    """Dump YAML and optionally write it to disk."""
    yaml_str = dump_yaml(data, flow_style=flow_style, width=width)
    if output_path:
        write_text(output_path, yaml_str)
    return yaml_str


def enrich_yaml_metadata(
    result: dict[str, Any], *, include_metadata: bool, flow_style: bool
) -> dict[str, Any]:
    """Add YAML-specific metadata while preserving JsonSerializer metadata layout."""
    if include_metadata:
        result["metadata"]["format"] = "yaraast-yaml"
        result["metadata"]["serializer"] = "YamlSerializer"
        result["metadata"]["yaml_features"] = {
            "flow_style": flow_style,
            "human_readable": True,
            "preserves_order": True,
        }
    return result



"""Real tests for YARA-X compatibility checker (no mocks)."""

from __future__ import annotations

from textwrap import dedent

from yaraast.parser import Parser
from yaraast.yarax.compatibility_checker import YaraXCompatibilityChecker
from yaraast.yarax.feature_flags import YaraXFeatures


def _parse_yara(code: str):
    return Parser().parse(dedent(code))


def test_compatibility_duplicate_modifiers() -> None:
    code = """
    private private rule dup_mods {
        condition:
            true
    }
    """
    ast = _parse_yara(code)
    checker = YaraXCompatibilityChecker(YaraXFeatures.yarax_strict())
    issues = checker.check(ast)

    assert any(i.issue_type == "duplicate_modifier" for i in issues)


def test_compatibility_base64_length() -> None:
    code = """
    rule short_base64 {
        strings:
            $a = "aa" base64
        condition:
            $a
    }
    """
    ast = _parse_yara(code)
    features = YaraXFeatures.yarax_strict()
    features.minimum_base64_length = 3
    checker = YaraXCompatibilityChecker(features)

    issues = checker.check(ast)
    assert any(i.issue_type == "base64_too_short" for i in issues)


def test_compatibility_regex_unescaped_brace() -> None:
    code = r"""
    rule bad_regex {
        strings:
            $a = /a{b/
        condition:
            $a
    }
    """
    ast = _parse_yara(code)
    checker = YaraXCompatibilityChecker(YaraXFeatures.yarax_strict())

    issues = checker.check(ast)
    assert any(i.issue_type == "unescaped_brace" for i in issues)



from __future__ import annotations

from yaraast.ast.base import YaraFile
from yaraast.ast.conditions import Condition, InExpression
from yaraast.ast.expressions import (
    BooleanLiteral,
    Identifier,
    IntegerLiteral,
    ParenthesesExpression,
    RangeExpression,
    SetExpression,
    StringLength,
    StringLiteral,
    StringOffset,
    UnaryExpression,
)
from yaraast.ast.meta import Meta
from yaraast.ast.modifiers import StringModifier
from yaraast.ast.modules import DictionaryAccess, ModuleReference
from yaraast.ast.operators import DefinedExpression, StringOperatorExpression
from yaraast.ast.rules import Import, Include, Rule, Tag
from yaraast.ast.strings import (
    HexAlternative,
    HexByte,
    HexNibble,
    HexString,
    PlainString,
    RegexString,
    StringDefinition,
)
from yaraast.codegen.advanced_generator import AdvancedCodeGenerator
from yaraast.codegen.formatting import BraceStyle, FormattingConfig, StringStyle
from yaraast.codegen.generator import CodeGenerator


def test_codegen_generator_additional_visit_paths() -> None:
    gen = CodeGenerator()

    yara_file = YaraFile(
        imports=[Import(module="pe", alias="p")],
        includes=[Include(path="common.yar")],
        rules=[
            Rule(
                name="r1",
                modifiers=["private"],
                tags=[Tag("t1")],
                meta=[Meta("author", "me"), object()],
                strings=[
                    PlainString(
                        "$a", value="x", modifiers=[StringModifier.from_name_value("ascii")]
                    ),
                    HexString("$h", tokens=[HexByte(0x4D), HexNibble(high=False, value=0xA)]),
                    RegexString("$r", regex="ab.*"),
                ],
                condition=BooleanLiteral(True),
            ),
        ],
    )
    out = gen.generate(yara_file)
    assert 'import "pe" as p' in out
    assert 'include "common.yar"' in out
    assert "private rule r1 : t1 {" in out

    assert gen.visit_string_definition(StringDefinition("$x")) == ""
    assert gen.visit_expression(Condition()) == ""
    assert gen.visit_condition(Condition()) == ""
    assert gen.visit_hex_token(HexByte(1)) == ""
    assert gen.visit_meta(Meta("k", "v")) == 'k = "v"'
    assert gen.visit_meta(Meta("k", False)) == "k = false"
    assert gen.visit_meta(Meta("k", 7)) == "k = 7"
    assert (
        gen.visit_dictionary_access(DictionaryAccess(ModuleReference("pe"), "CompanyName"))
        == 'pe["CompanyName"]'
    )
    assert (
        gen.visit_dictionary_access(DictionaryAccess(ModuleReference("pe"), IntegerLiteral(1)))
        == "pe[1]"
    )
    assert gen.visit_defined_expression(DefinedExpression(Identifier("$a"))) == "defined $a"
    assert (
        gen.visit_string_operator_expression(
            StringOperatorExpression(StringLiteral("a"), "icontains", StringLiteral("b")),
        )
        == '"a" icontains "b"'
    )
    assert (
        gen.visit_in_expression(InExpression(subject="$a", range=Identifier("filesize")))
        == "$a in filesize"
    )
    assert (
        gen.visit_hex_alternative(HexAlternative([[HexByte(0x4D)], [HexByte(0x5A)]]))
        == "( 4D | 5A )"
    )
    assert gen.visit_string_offset(StringOffset(string_id="a", index=None)) == "@a"
    assert gen.visit_string_length(StringLength(string_id="a", index=None)) == "!a"
    assert gen.visit_unary_expression(UnaryExpression("-", IntegerLiteral(1))) == "-1"
    assert (
        gen.visit_in_expression(
            InExpression(
                subject="$a",
                range=ParenthesesExpression(RangeExpression(IntegerLiteral(1), IntegerLiteral(3))),
            )
        )
        == "$a in (1..3)"
    )


def test_advanced_generator_additional_paths() -> None:
    cfg = FormattingConfig(
        sort_meta=True,
        string_style=StringStyle.COMPACT,
        space_before_colon=True,
        space_after_colon=False,
        sort_rules=False,
        max_line_length=4,
    )
    adv = AdvancedCodeGenerator(cfg)

    rule_with_meta = Rule(
        name="a_rule",
        modifiers=["private"],
        tags=["x"],
        meta=[Meta("z", "quoted"), Meta("missing", "")],
        strings=[
            PlainString("$a", value="x"),
            HexString("$h", tokens=[HexByte(0x4D)]),
            RegexString("$r", regex="re"),
        ],
        condition=BooleanLiteral(True),
    )
    rule_without_meta = Rule(name="b_rule", condition=BooleanLiteral(False))
    out = adv.generate(
        YaraFile(
            imports=[Import("pe")],
            includes=[Include("c.yar")],
            rules=[rule_without_meta, rule_with_meta],
        )
    )

    assert 'import "pe"' in out
    assert 'include "c.yar"' in out
    assert "rule a_rule :x" in out
    assert '$a="x"' in out
    assert "$h={ 4d }" in out or "$h = { 4d }" in out or "$h={ 4D }" in out or "$h = { 4D }" in out
    assert "$r=/re/" in out

    adv2 = AdvancedCodeGenerator(
        FormattingConfig(space_around_operators=False, space_after_comma=False)
    )
    adv2.generate(YaraFile(rules=[]))
    assert "(" in adv2.visit_set_expression(
        SetExpression(elements=[IntegerLiteral(1), IntegerLiteral(2)])
    )


def test_advanced_generator_additional_formatting_paths() -> None:
    new_line_cfg = FormattingConfig(
        brace_style=BraceStyle.NEW_LINE, string_style=StringStyle.TABULAR
    )
    adv = AdvancedCodeGenerator(new_line_cfg)
    rule = Rule(
        name="fmt",
        tags=[Tag("one"), Tag("two")],
        meta={"b": 2, "a": '"quoted"', "flag": True},
        strings=[
            PlainString("$b", value="x", modifiers=[StringModifier.from_name_value("ascii")]),
            PlainString("$a", value="y"),
        ],
        condition=BooleanLiteral(True),
    )
    out = adv.generate(YaraFile(rules=[rule]))
    assert "rule fmt" in out
    assert "{\n" in out
    assert 'a    = "quoted"' in out or 'a = "quoted"' in out
    assert "flag = true" in out
    assert "$a" in out and "$b" in out


def test_advanced_generator_direct_remaining_branches() -> None:
    adv = AdvancedCodeGenerator(
        FormattingConfig(
            string_style=StringStyle.ALIGNED,
            align_string_modifiers=False,
            blank_lines_between_sections=2,
            space_after_comma=True,
            max_line_length=999,
        )
    )

    invalid_meta_list = [Meta("k", ""), object()]
    adv._write_meta_section(invalid_meta_list)
    meta_output = adv.buffer.getvalue()
    assert 'k = ""' in meta_output

    adv.buffer.seek(0)
    adv.buffer.truncate(0)
    assert adv._get_max_key_length([]) == 0
    adv._string_definitions = []
    adv._write_aligned_strings()
    assert adv.buffer.getvalue() == ""

    plain = PlainString(
        "$a",
        value="x",
        modifiers=[StringModifier.from_name_value("ascii")],
    )
    hexs = HexString(
        "$h",
        tokens=[HexByte(0x4D)],
        modifiers=[StringModifier.from_name_value("wide")],
    )
    regex = RegexString(
        "$r",
        regex="abc",
        modifiers=[StringModifier.from_name_value("nocase")],
    )
    rule = Rule(
        name="styled",
        meta=[Meta("author", "me"), object()],
        strings=[plain, hexs, regex],
        condition=BooleanLiteral(True),
    )
    out = adv.generate(YaraFile(rules=[rule]))
    assert "\n\nmeta:\n" not in out
    assert "\n\n\n    strings:\n" in out
    assert '$a = "x" ascii' in out
    assert "$h = { 4d } wide" in out or "$h = { 4D } wide" in out
    assert "$r = /abc/ nocase" in out

    adv2 = AdvancedCodeGenerator(FormattingConfig(space_after_comma=False))
    expr_out = adv2.visit_set_expression(
        SetExpression(elements=[IntegerLiteral(1), IntegerLiteral(2)])
    )
    assert expr_out.endswith("(1,2)")

    adv3 = AdvancedCodeGenerator(FormattingConfig())
    assert adv3._format_hex_token(HexByte(0x4D)) in {"4d", "4D"}


def test_advanced_generator_final_remaining_string_and_section_paths() -> None:
    class FlakyMeta:
        def __init__(self) -> None:
            self._has_key = True

        @property
        def key(self) -> str:
            if self._has_key:
                self._has_key = False
                return "ephemeral"
            raise AttributeError("key disappeared")

    adv = AdvancedCodeGenerator(
        FormattingConfig(
            section_order=["strings", "meta", "condition"],
            blank_lines_between_sections=1,
            string_style=StringStyle.ALIGNED,
            space_after_comma=True,
        )
    )

    # Exercise the defensive continue in _write_meta_section when a key disappears.
    adv._write_meta_section([FlakyMeta()])
    assert "meta:" in adv.buffer.getvalue()

    adv.buffer.seek(0)
    adv.buffer.truncate(0)

    rule = Rule(
        name="ordered",
        meta=[Meta("author", "me")],
        strings=[
            PlainString("$a", value="x", modifiers=[StringModifier.from_name_value("ascii")]),
            HexString(
                "$h", tokens=[HexByte(0x4D)], modifiers=[StringModifier.from_name_value("wide")]
            ),
            RegexString("$r", regex="abc", modifiers=[StringModifier.from_name_value("nocase")]),
        ],
        condition=BooleanLiteral(True),
    )
    out = adv.generate(YaraFile(rules=[rule]))
    assert "\n\n    meta:\n" in out

    adv2 = AdvancedCodeGenerator(
        FormattingConfig(string_style=StringStyle.ALIGNED, space_after_comma=True)
    )
    adv2.generate(YaraFile(rules=[]))
    assert (
        adv2.visit_plain_string(
            PlainString("$a", value="x", modifiers=[StringModifier.from_name_value("ascii")])
        )
        == ""
    )
    plain_output = adv2.buffer.getvalue()
    assert plain_output == '$a = "x" '

    adv2.buffer.seek(0)
    adv2.buffer.truncate(0)
    assert (
        adv2.visit_hex_string(
            HexString(
                "$h", tokens=[HexByte(0x4D)], modifiers=[StringModifier.from_name_value("wide")]
            )
        )
        == ""
    )
    hex_output = adv2.buffer.getvalue()
    assert hex_output in {"$h = { 4d } ", "$h = { 4D } "}

    adv2.buffer.seek(0)
    adv2.buffer.truncate(0)
    assert (
        adv2.visit_regex_string(
            RegexString("$r", regex="abc", modifiers=[StringModifier.from_name_value("nocase")])
        )
        == ""
    )
    assert adv2.buffer.getvalue() == "$r = /abc/ "

    adv3 = AdvancedCodeGenerator(FormattingConfig(space_after_comma=True))
    assert adv3.visit_set_expression(
        SetExpression(elements=[IntegerLiteral(1), IntegerLiteral(2)])
    ).endswith("(1, 2)")



"""Additional tests for performance service helpers."""

from __future__ import annotations

from pathlib import Path
from types import SimpleNamespace

from yaraast.cli import performance_services as ps
from yaraast.parser import Parser
from yaraast.performance.streaming_parser import StreamingParser


def _ast(name: str = "r"):
    return Parser().parse(f"rule {name} {{ condition: true }}")


def test_get_parse_iterator_collect_paths_and_stream_summary(tmp_path: Path) -> None:
    parser = StreamingParser()
    file_path = tmp_path / "one.yar"
    file_path.write_text("rule one { condition: true }\n", encoding="utf-8")

    split_iter = list(ps.get_parse_iterator(parser, file_path, True, "*.yar", False))
    assert split_iter and split_iter[0].rule_name == "one"

    parse_iter = list(ps.get_parse_iterator(parser, file_path, False, "*.yar", False))
    assert parse_iter and parse_iter[0].rule_name == "one"

    dir_iter = list(ps.get_parse_iterator(parser, tmp_path, False, "*.yar", False))
    assert dir_iter and dir_iter[0].rule_name == "one"

    missing = tmp_path / "missing"
    paths = ps.collect_file_paths((str(file_path), str(tmp_path), str(missing)))
    assert file_path in paths
    assert all(path.exists() for path in paths)

    summary = ps.summarize_stream_results(
        [
            SimpleNamespace(status=SimpleNamespace(value="success")),
            SimpleNamespace(status=SimpleNamespace(value="error")),
            SimpleNamespace(status=SimpleNamespace(value="other")),
        ]
    )
    assert len(summary["successful"]) == 1
    assert len(summary["failed"]) == 1


def test_extract_successful_asts_and_file_name_mapping_paths(tmp_path: Path) -> None:
    file_paths = [tmp_path / "a.yar"]
    ok_ast = _ast("ok")
    bad_ast = SimpleNamespace(_parse_error=True)
    job = SimpleNamespace(
        status=SimpleNamespace(value="completed"), result=[ok_ast, bad_ast, ok_ast]
    )
    skipped_job = SimpleNamespace(status=SimpleNamespace(value="failed"), result=[ok_ast])

    asts, names = ps.extract_successful_asts([job, skipped_job], file_paths, chunk_size=2)
    assert asts == [ok_ast, ok_ast]
    assert names == [str(file_paths[0])]

    assert ps._get_corresponding_file_name(1, 0, file_paths, 2) is None


def test_build_parallel_summary_and_plans_cover_remaining_branches() -> None:
    summary = ps.build_parallel_summary(
        [Path("a.yar")],
        [_ast("a")],
        {"avg_time_per_rule": 0.5, "jobs_completed": 1, "max_workers": 4},
        total_time=0,
    )
    assert summary["avg_job_time"] == 0.5
    assert summary["speedup"] == 1
    assert summary["workers_used"] == 4

    summary2 = ps.build_parallel_summary(
        [Path("a.yar")] * 6,
        [_ast("a")] * 6,
        {
            "avg_time_per_rule": 1.5,
            "jobs_completed": 6,
            "total_processing_time": 9.0,
            "max_workers": 4,
        },
        total_time=3.0,
    )
    assert summary2["speedup"] == 3.0  # 6 jobs * 1.5s avg = 9s sequential / 3s parallel

    summary3 = ps.build_parallel_summary(
        [Path("a.yar")],
        [],
        {"avg_time_per_rule": 0.0, "jobs_completed": 0, "max_workers": 2},
        total_time=5.0,
    )
    assert summary3["speedup"] == 1.0

    assert ps._build_strategy_messages(500) == [
        "Use batch processing with moderate parallelism",
        "Enable object pooling",
        "Monitor memory usage",
    ]
    assert ps._build_strategy_messages(5000) == [
        "Use streaming parser with small batches",
        "Enable aggressive memory management",
        "Consider distributed processing",
    ]
    assert ps._build_memory_plan(10, None, {"memory_limit_mb": 64}) is None
    assert ps._build_time_plan(10, None) is None


def test_run_parallel_analysis_and_build_output_data_more_paths(tmp_path: Path) -> None:
    file_path = tmp_path / "one.yar"
    file_path.write_text("rule one { condition: true }\n", encoding="utf-8")
    output_dir = tmp_path / "out"

    run_results, total_time = ps.run_parallel_analysis(
        [file_path],
        max_workers=1,
        chunk_size=1,
        analysis_type="all",
        output_dir=output_dir,
    )
    assert total_time >= 0
    assert len(run_results["successful_asts"]) == 1
    assert run_results["complexity_results"]
    assert run_results["dependency_graphs"]

    result_obj = SimpleNamespace(
        file_path=str(file_path),
        rule_name="one",
        status=SimpleNamespace(value="success"),
        error=None,
        parse_time=0.1,
        rule_count=1,
        import_count=0,
    )
    out = ps.build_stream_output_data([result_obj], [result_obj], [], 0.5, {"peak_memory_mb": 1})
    assert out["summary"]["success_rate"] == 100.0
    assert out["results"][0]["status"] == "success"



from __future__ import annotations

from yaraast.yaral.ast_nodes import (
    ArithmeticExpression,
    ConditionalExpression,
    EventExistsCondition,
)
from yaraast.yaral.validator import YaraLValidator


def test_validator_conditional_expression_without_accept_condition() -> None:
    validator = YaraLValidator()
    validator.current_rule = "r6"

    validator.visit_yaral_conditional_expression(
        ConditionalExpression(condition=True, true_value=1, false_value=0)
    )

    assert validator.errors == []
    assert validator.warnings == []
    assert validator.used_events == set()


def test_validator_arithmetic_expression_partial_and_non_visitable_operands() -> None:
    validator = YaraLValidator()
    validator.current_rule = "r7"

    validator.visit_yaral_arithmetic_expression(
        ArithmeticExpression(
            operator="+",
            left=EventExistsCondition(event="left_only"),
            right=1,
        )
    )
    assert "left_only" in validator.used_events

    validator.used_events.clear()
    validator.visit_yaral_arithmetic_expression(
        ArithmeticExpression(
            operator="+",
            left=1,
            right=2,
        )
    )
    assert validator.used_events == set()



from __future__ import annotations

from yaraast.lexer.tokens import TokenType as T
from yaraast.yaral.ast_nodes import EventAssignment, JoinCondition
from yaraast.yaral.enhanced_parser import EnhancedYaraLParser
from yaraast.yaral.lexer import YaraLToken
from yaraast.yaral.tokens import YaraLTokenType


def _tok(tt: T, value, yt: YaraLTokenType | None = None) -> YaraLToken:
    return YaraLToken(type=tt, value=value, line=1, column=1, length=1, yaral_type=yt)


def _set_tokens(p: EnhancedYaraLParser, toks: list[YaraLToken]) -> None:
    p.tokens = [*toks, _tok(T.EOF, None, YaraLTokenType.EOF)]
    p.current = 0


def test_enhanced_events_section_handles_join_and_garbage() -> None:
    p = EnhancedYaraLParser("")
    _set_tokens(
        p,
        [
            _tok(T.IDENTIFIER, "events"),
            _tok(T.COLON, ":"),
            _tok(T.IDENTIFIER, "join"),
            _tok(T.IDENTIFIER, "left"),
            _tok(T.IDENTIFIER, "on"),
            _tok(T.IDENTIFIER, "$x", YaraLTokenType.EVENT_VAR),
            _tok(T.IDENTIFIER, "with"),
            _tok(T.IDENTIFIER, "right"),
            _tok(T.PLUS, "+"),
            _tok(T.RBRACE, "}"),
        ],
    )

    section = p._parse_events_section()
    assert len(section.statements) == 1
    assert isinstance(section.statements[0], JoinCondition)


def test_enhanced_event_statement_none_and_identifier_path_forms() -> None:
    p = EnhancedYaraLParser("")
    _set_tokens(p, [_tok(T.IDENTIFIER, "$e", YaraLTokenType.EVENT_VAR)])
    assert p._parse_event_statement() is None

    p2 = EnhancedYaraLParser("")
    _set_tokens(
        p2,
        [
            _tok(T.IDENTIFIER, "$e", YaraLTokenType.EVENT_VAR),
            _tok(T.IDENTIFIER, "metadata"),
            _tok(T.EQ, "="),
            _tok(T.STRING, "LOGIN"),
        ],
    )
    stmt = p2._parse_event_statement()
    assert isinstance(stmt, list)
    assert len(stmt) == 1
    assert isinstance(stmt[0], EventAssignment)
    assert stmt[0].field_path.parts == ["metadata"]


def test_enhanced_event_statement_breaks_when_event_missing_after_loop_entry() -> None:
    p = EnhancedYaraLParser("")
    _set_tokens(
        p,
        [
            _tok(T.IDENTIFIER, "metadata"),
            _tok(T.EQ, "="),
            _tok(T.STRING, "LOGIN"),
        ],
    )
    assert p._parse_event_statement() is None


def test_enhanced_complex_patterns_and_join_condition_delegate() -> None:
    p = EnhancedYaraLParser("")
    _set_tokens(p, [_tok(T.IDENTIFIER, "nope")])
    assert p._parse_complex_event_pattern() is None

    p2 = EnhancedYaraLParser("")
    _set_tokens(p2, [_tok(T.IDENTIFIER, "$e", YaraLTokenType.EVENT_VAR)])
    condition = p2._parse_join_condition()
    assert condition.__class__.__name__ == "EventExistsCondition"



"""Operational workflows for the metrics subsystem."""

from __future__ import annotations

from collections.abc import Callable
from dataclasses import dataclass
from pathlib import Path
from typing import Any

from yaraast.ast.base import YaraFile
from yaraast.errors import ValidationError
from yaraast.metrics.capabilities import get_capability
from yaraast.metrics.complexity_model import ComplexityMetrics
from yaraast.metrics.facade import METRICS
from yaraast.metrics.html_tree import HtmlTreeGenerator
from yaraast.metrics.string_diagrams import StringDiagramGenerator

try:
    from yaraast.metrics.dependency_graph import DependencyGraphGenerator
except ModuleNotFoundError:
    DependencyGraphGenerator = None  # type: ignore[assignment]


@dataclass
class MetricsReportData:
    base_name: str
    complexity_metrics: ComplexityMetrics
    complexity_payload: dict[str, Any]
    generated_files: list[str]


def analyze_complexity(ast: YaraFile) -> ComplexityMetrics:
    analyzer = METRICS.new_complexity_analyzer()
    return analyzer.analyze(ast)


def is_graphviz_error(error: Exception) -> bool:
    """Check if error is caused by missing graphviz installation."""
    # Check exception type first (more reliable than string matching)
    error_type = type(error).__name__
    if error_type in ("ExecutableNotFound", "CalledProcessError"):
        return True
    error_str = str(error)
    graphviz_indicators = [
        "dot",
        "graphviz",
        "ExecutableNotFound",
        "No such file or directory",
    ]
    return any(indicator.lower() in error_str.lower() for indicator in graphviz_indicators)


def build_complexity_payload(metrics: ComplexityMetrics) -> dict[str, Any]:
    capability = get_capability("complexity")
    result = metrics.to_dict()
    result["quality_score"] = metrics.get_quality_score()
    result["quality_grade"] = metrics.get_complexity_grade()
    result["heuristic"] = True
    result["analysis_kind"] = "heuristic"
    if capability is not None:
        result["capability"] = capability.name
        result["capability_outputs"] = list(capability.outputs)
    return result


def generate_dependency_graphs(
    ast: YaraFile,
    output_dir: Path,
    base_name: str,
    image_format: str,
    generator_factory: Callable[[], Any] | None = DependencyGraphGenerator,
) -> list[str]:
    if generator_factory is None:
        msg = "Graph visualization requires the 'graphviz' Python package."
        raise RuntimeError(msg)
    generator = (
        generator_factory()
        if generator_factory is not DependencyGraphGenerator
        else METRICS.new_dependency_graph_generator()
    )
    generated = []

    deps_path = output_dir / f"{base_name}_dependencies.{image_format}"
    generator.generate_graph(ast, str(deps_path), image_format)
    generated.append(deps_path.name)

    rules_path = output_dir / f"{base_name}_rules.{image_format}"
    generator.generate_rule_graph(ast, str(rules_path), image_format)
    generated.append(rules_path.name)

    complexity_path = output_dir / f"{base_name}_complexity_graph.{image_format}"
    metrics = analyze_complexity(ast)
    generator.generate_complexity_graph(
        ast, metrics.cyclomatic_complexity, str(complexity_path), image_format
    )
    generated.append(complexity_path.name)
    return generated


def generate_html_tree(
    ast: YaraFile,
    output_dir: Path,
    base_name: str,
    interactive: bool = True,
    generator_factory: Callable[[], Any] = HtmlTreeGenerator,
) -> str:
    generator = (
        generator_factory()
        if generator_factory is not HtmlTreeGenerator
        else METRICS.new_html_tree_generator()
    )
    output_path = output_dir / f"{base_name}_tree.html"
    if interactive:
        generator.generate_interactive_html(ast, str(output_path))
    else:
        generator.generate_html(ast, str(output_path))
    return output_path.name


def generate_pattern_diagrams(
    ast: YaraFile,
    output_dir: Path,
    base_name: str,
    image_format: str,
    generator_factory: Callable[[], Any] = StringDiagramGenerator,
) -> list[str]:
    generator = (
        generator_factory()
        if generator_factory is not StringDiagramGenerator
        else METRICS.new_string_diagram_generator()
    )
    generated = []

    flow_path = output_dir / f"{base_name}_pattern_flow.{image_format}"
    generator.generate_pattern_flow_diagram(ast, str(flow_path), image_format)
    generated.append(flow_path.name)

    complexity_path = output_dir / f"{base_name}_pattern_complexity.{image_format}"
    generator.generate_pattern_complexity_diagram(ast, str(complexity_path), image_format)
    generated.append(complexity_path.name)

    hex_path = output_dir / f"{base_name}_hex_patterns.{image_format}"
    generator.generate_hex_pattern_diagram(ast, str(hex_path), image_format)
    generated.append(hex_path.name)
    return generated


def build_report(
    ast: YaraFile, output_dir: Path, base_name: str, image_format: str
) -> MetricsReportData:
    metrics = analyze_complexity(ast)
    payload = build_complexity_payload(metrics)

    generated_files = []
    try:
        generated_files.extend(generate_dependency_graphs(ast, output_dir, base_name, image_format))
    except Exception as exc:  # suppress missing graphviz, re-raise others
        if not is_graphviz_error(exc):
            raise
    generated_files.append(generate_html_tree(ast, output_dir, base_name, interactive=True))
    try:
        generated_files.extend(generate_pattern_diagrams(ast, output_dir, base_name, image_format))
    except Exception as exc:  # suppress missing graphviz, re-raise others
        if not is_graphviz_error(exc):
            raise

    return MetricsReportData(
        base_name=base_name,
        complexity_metrics=metrics,
        complexity_payload=payload,
        generated_files=generated_files,
    )


def determine_graph_output_path(
    yara_file: str, output: str | None, graph_type: str, fmt: str
) -> str:
    if output:
        return output
    base_name = Path(yara_file).stem
    return f"{base_name}_graph_{graph_type}.{fmt}"


def generate_dependency_graph(
    ast: YaraFile, graph_type: str, output_path: str, fmt: str, engine: str
) -> str:
    if DependencyGraphGenerator is None:
        msg = "Graph visualization requires the 'graphviz' Python package."
        raise RuntimeError(msg)
    generator = METRICS.new_dependency_graph_generator()
    return generate_dependency_graph_with_generator(
        generator, ast, graph_type, output_path, fmt, engine
    )


def generate_dependency_graph_with_generator(
    generator: Any,
    ast: YaraFile,
    graph_type: str,
    output_path: str,
    fmt: str,
    engine: str,
) -> tuple[str, Any]:
    if graph_type == "full":
        return generator.generate_graph(ast, output_path, fmt, engine), generator
    if graph_type == "rules":
        return generator.generate_rule_graph(ast, output_path, fmt), generator
    if graph_type == "modules":
        return generator.generate_module_graph(ast, output_path, fmt), generator
    if graph_type == "complexity":
        metrics = analyze_complexity(ast)
        return (
            generator.generate_complexity_graph(
                ast, metrics.cyclomatic_complexity, output_path, fmt
            ),
            generator,
        )
    raise ValidationError(f"Unknown graph type: {graph_type}")


def determine_pattern_output_path(
    yara_file: str, output: str | None, pattern_type: str, fmt: str
) -> str:
    if output:
        return output
    base_name = Path(yara_file).stem
    return f"{base_name}_patterns_{pattern_type}.{fmt}"


def generate_pattern_diagram(ast: YaraFile, pattern_type: str, output_path: str, fmt: str) -> str:
    generator = METRICS.new_string_diagram_generator()
    return generate_pattern_diagram_with_generator(generator, ast, pattern_type, output_path, fmt)


def generate_pattern_diagram_with_generator(
    generator: Any,
    ast: YaraFile,
    pattern_type: str,
    output_path: str,
    fmt: str,
) -> str:
    if pattern_type == "flow":
        return generator.generate_pattern_flow_diagram(ast, output_path, fmt)
    if pattern_type == "complexity":
        return generator.generate_pattern_complexity_diagram(ast, output_path, fmt)
    if pattern_type == "similarity":
        return generator.generate_pattern_similarity_diagram(ast, output_path, fmt)
    if pattern_type == "hex":
        return generator.generate_hex_pattern_diagram(ast, output_path, fmt)
    raise ValidationError(f"Unknown pattern type: {pattern_type}")


def generate_html_tree_file(
    ast: YaraFile,
    output_path: str,
    title: str,
    interactive: bool,
    include_metadata: bool,
) -> str:
    generator = HtmlTreeGenerator(include_metadata=include_metadata)
    if interactive:
        generator.generate_interactive_html(ast, output_path, title)
    else:
        generator.generate_html(ast, output_path, title)
    return output_path



"""Extra real coverage for condition builder, hex builder, and HTML tree nodes."""

from __future__ import annotations

import pytest

from yaraast.ast.expressions import Identifier
from yaraast.builder.condition_builder import ConditionBuilder
from yaraast.builder.hex_string_builder import HexStringBuilder
from yaraast.errors import ValidationError
from yaraast.metrics.html_tree import HtmlTreeGenerator


def test_condition_builder_n_of_them_and_empty_arithmetic_error() -> None:
    expr = ConditionBuilder().n_of(2, "them").build()
    assert isinstance(expr.string_set, Identifier)
    assert expr.string_set.name == "them"

    with pytest.raises(ValidationError, match="Cannot apply \\+ to empty expression"):
        ConditionBuilder().add(1)


def test_hex_string_builder_invalid_low_nibble_and_unknown_pattern_part() -> None:
    with pytest.raises(ValidationError, match="Invalid nibble pattern: \\?G"):
        HexStringBuilder().nibble("?G")

    builder = HexStringBuilder()
    builder.pattern("XYZ")
    assert builder.build() == []


def test_html_tree_nodes_condition_section_none() -> None:
    gen = HtmlTreeGenerator()
    assert gen._condition_section(None) is None


def test_html_tree_nodes_write_meta_and_string_helpers(tmp_path) -> None:
    gen = HtmlTreeGenerator()
    out = tmp_path / "tree.html"
    gen._write_output(str(out), "<html>ok</html>")
    assert out.read_text(encoding="utf-8") == "<html>ok</html>"

    meta = gen._meta_section({"author": "unit", "score": 5})
    assert meta is not None
    assert meta["label"] == "Meta"
    assert len(meta["children"]) == 2

    string_node = gen._string_node(
        "Plain String: $a",
        value='"abc"',
        details="3 bytes",
        children=[{"id": "node_x", "label": "child", "node_class": "child"}],
    )
    assert string_node["details"] == "3 bytes"
    assert string_node["children"][0]["label"] == "child"



"""Rename provider for YARA Language Server."""

from __future__ import annotations

from lsprotocol.types import Position, Range, WorkspaceEdit

from yaraast.lsp.runtime import DocumentContext, LspRuntime


class RenameProvider:
    """Provides rename functionality."""

    def __init__(self, runtime: LspRuntime | None = None) -> None:
        self.runtime = runtime

    def prepare_rename(self, text: str, position: Position, uri: str | None = None) -> Range | None:
        """
        Prepare for rename operation.

        Args:
            text: The YARA source code
            position: The cursor position

        Returns:
            Range of the symbol to rename or None if not renameable
        """
        resolved = (
            self.runtime.resolve_symbol(uri, text, position)
            if self.runtime and uri
            else DocumentContext(uri or "file://local.yar", text).resolve_symbol(position)
        )
        if resolved is not None and resolved.kind in {"string", "rule"}:
            return resolved.range
        return None

    def rename(
        self,
        text: str,
        position: Position,
        new_name: str,
        uri: str,
    ) -> WorkspaceEdit | None:
        """
        Perform rename operation.

        Args:
            text: The YARA source code
            position: The position of the symbol to rename
            new_name: The new name
            uri: The document URI

        Returns:
            WorkspaceEdit with all the changes
        """
        doc = (
            self.runtime.ensure_document(uri, text)
            if self.runtime and uri
            else DocumentContext(uri, text)
        )
        resolved = (
            self.runtime.resolve_symbol(uri, text, position)
            if self.runtime and uri
            else doc.resolve_symbol(position)
        )
        if resolved is None:
            return None

        # Handle string identifier renaming
        if resolved.kind == "string":
            base_identifier = resolved.normalized_name
            edits = doc.build_string_rename_edits(base_identifier, new_name)
            if edits:
                return WorkspaceEdit(changes={uri: edits})
            return None

        # Handle rule name renaming
        if resolved.kind == "rule":
            if self.runtime:
                changes = self.runtime.rename_rule(resolved.normalized_name, new_name)
                if changes:
                    return WorkspaceEdit(changes=changes)
            edits = doc.rename_rule_edits(resolved.normalized_name, new_name)
            if edits:
                return WorkspaceEdit(changes={uri: edits})

        return None

    def _is_rule_name(self, text: str, position: Position) -> bool:
        """Check if position is at a rule name."""
        lines = text.split("\n")
        if position.line >= len(lines):
            return False

        line = lines[position.line]
        return "rule" in line and "{" not in line



"""More tests for LSP hover provider (no mocks)."""

from __future__ import annotations

from lsprotocol.types import Position

from yaraast.lsp.hover import HoverProvider


def _pos(line: int, char: int) -> Position:
    return Position(line=line, character=char)


def test_hover_keyword_and_builtin() -> None:
    text = "rule a { condition: uint16(0) and true }"
    provider = HoverProvider()

    hover_rule = provider.get_hover(text, _pos(0, 1))
    assert hover_rule is not None
    assert "keyword" in hover_rule.contents.value

    hover_uint = provider.get_hover(text, _pos(0, 22))
    assert hover_uint is not None
    assert "built-in function" in hover_uint.contents.value


def test_hover_string_identifier() -> None:
    text = """
rule a {
  strings:
    $a = "x" ascii
  condition:
    $a
}
""".lstrip()

    provider = HoverProvider()
    hover = provider.get_hover(text, _pos(4, 5))
    assert hover is not None
    assert "$a" in hover.contents.value



"""Additional tests for comment-preserving lexer and resolution dependency graph."""

from __future__ import annotations

from pathlib import Path

from yaraast.ast.base import YaraFile
from yaraast.ast.rules import Import, Include, Rule
from yaraast.lexer.comment_preserving_lexer import CommentPreservingLexer
from yaraast.lexer.tokens import TokenType
from yaraast.resolution.dependency_graph import DependencyGraph, DependencyNode


def test_comment_preserving_lexer_helpers_and_preserve_toggle() -> None:
    lexer = CommentPreservingLexer("// first\nrule a { /* block */ condition: true }")
    tokens = lexer.tokenize()

    comments = [t for t in tokens if t.type == TokenType.COMMENT]
    assert len(comments) == 2
    assert lexer.get_comments() == comments

    line_comment = CommentPreservingLexer("// line")
    token = line_comment._read_line_comment()
    assert token is not None
    assert token.type == TokenType.COMMENT
    assert token.value == "// line"

    block_comment = CommentPreservingLexer("/* block */")
    token = block_comment._read_block_comment()
    assert token is not None
    assert token.type == TokenType.COMMENT
    assert token.value == "/* block */"

    unterminated = CommentPreservingLexer("/* block")
    token = unterminated._read_block_comment()
    assert token is not None
    assert token.value == "/* block"

    multiline = CommentPreservingLexer("/* one\nline */rule c { condition: true }")
    tokens = multiline.tokenize()
    comments = [t for t in tokens if t.type == TokenType.COMMENT]
    assert comments[0].line == 1
    assert comments[0].column == 1

    no_preserve = CommentPreservingLexer("// c1\n/* c2 */rule b { condition: true }")
    no_preserve.set_preserve_comments(False)
    tokens = no_preserve.tokenize()
    assert all(t.type != TokenType.COMMENT for t in tokens)
    assert no_preserve.get_comments() == []

    no_preserve.set_preserve_comments(True)
    assert no_preserve.preserve_comments is True

    lexer.clear_comments()
    assert lexer.get_comments() == []


def test_dependency_graph_transitive_queries_cycles_and_export() -> None:
    graph = DependencyGraph()
    file_path = Path("file.yar")
    ast = YaraFile(
        imports=[Import(module="pe")],
        includes=[Include(path="shared.yar")],
        rules=[Rule(name='rule:"x"', modifiers=["private"], tags=[])],
    )
    graph.add_file(file_path, ast)

    assert str(file_path) in graph.nodes
    assert graph.get_rule_dependencies('rule:"x"') == set()
    assert graph.get_rule_dependencies("missing") == set()

    graph.nodes["shared.yar"] = DependencyNode("shared.yar", "file")
    graph.nodes["shared.yar"].dependents.add(str(file_path))
    graph.nodes['rule:rule:"x"'].dependencies.add("shared.yar")
    graph.nodes["shared.yar"].dependents.add('rule:rule:"x"')

    deps = graph.get_file_dependencies(str(file_path))
    assert "pe" in deps
    assert "shared.yar" in deps
    assert 'rule:rule:"x"' in deps

    dependents = graph.get_file_dependents("shared.yar")
    assert str(file_path) in dependents or 'rule:rule:"x"' in dependents

    graph.nodes["isolated"] = DependencyNode("isolated", "mystery")
    assert "isolated" in graph.get_isolated_nodes()

    cycle_graph = DependencyGraph()
    cycle_graph.nodes["A"] = DependencyNode("A", "file", dependencies={"B"})
    cycle_graph.nodes["B"] = DependencyNode("B", "file", dependencies={"A"})
    cycles = cycle_graph.find_cycles()
    assert cycles

    stats = graph.get_statistics()
    assert stats["file_count"] >= 1
    assert stats["rule_count"] == 1
    assert stats["module_count"] == 1

    dot = graph.export_dot()
    assert 'label="rule:\\"x\\""' in dot
    assert "shape=folder" in dot
    assert "shape=component" in dot
    assert '"isolated" [label="isolated",' in dot

    resolved = graph.get_file_dependencies(".")
    assert isinstance(resolved, set)
    assert graph.get_file_dependents("missing") == set()



from __future__ import annotations

from yaraast.ast.base import YaraFile
from yaraast.ast.expressions import (
    BinaryExpression,
    BooleanLiteral,
    DoubleLiteral,
    Identifier,
    IntegerLiteral,
    StringIdentifier,
    StringLiteral,
    StringWildcard,
    UnaryExpression,
)
from yaraast.ast.modifiers import MetaEntry
from yaraast.ast.rules import Import, Include, Rule, Tag
from yaraast.ast.strings import HexByte, HexString, PlainString, RegexString
from yaraast.performance.memory_optimizer import MemoryOptimizer, MemoryOptimizerTransformer


def test_memory_optimizer_rule_list_usage_and_batch_paths() -> None:
    optimizer = MemoryOptimizer(aggressive=True, gc_threshold=2)
    rule = Rule(
        name="dup",
        tags=[Tag("tag"), Tag("tag")],
        meta=[MetaEntry("author", "alice"), MetaEntry("author", "alice")],
        strings=[
            PlainString(identifier="$a", value="same"),
            RegexString(identifier="$r", regex="same"),
            HexString(identifier="$h", tokens=[HexByte(0x41)]),
        ],
        condition=BinaryExpression(
            left=Identifier("same"),
            operator="and",
            right=UnaryExpression("not", StringIdentifier("$a")),
        ),
    )

    optimized_rule = optimizer.optimize_rule(rule)
    assert optimized_rule is not rule  # optimizer returns a new copy

    optimized_rules = optimizer.optimize_rules([rule])
    assert optimized_rules == [rule]

    usage = optimizer.get_memory_usage()
    assert {"rss_mb", "vms_mb", "percent", "available_mb"} <= set(usage)

    batches = list(
        optimizer.batch_process_with_memory_limit(list(range(5)), lambda x: x + 1, batch_size=2)
    )
    assert batches == [[1, 2], [3, 4], [5]]


def test_memory_optimizer_transformer_visits_real_nodes() -> None:
    pool: dict[str, str] = {}
    transformer = MemoryOptimizerTransformer(pool, aggressive=True)

    yara_file = YaraFile(
        imports=[Import(module="pe")],
        includes=[Include(path="common.yar")],
        rules=[
            Rule(
                name="dup",
                tags=[Tag("dup")],
                meta=[MetaEntry("k", "dup")],
                strings=[
                    PlainString(identifier="$a", value="dup"),
                    HexString(identifier="$h", tokens=[HexByte(0x41)]),
                    RegexString(identifier="$r", regex="dup"),
                ],
                condition=BinaryExpression(
                    left=StringLiteral("dup"),
                    operator="or",
                    right=Identifier("dup"),
                ),
            ),
        ],
    )

    optimized = transformer.visit(yara_file)
    assert optimized is not yara_file  # returns a new copy
    assert transformer.nodes_processed >= 1
    assert optimized.rules[0].location is None
    assert optimized.rules[0].name == "dup"
    assert optimized.rules[0].strings[0].value == "dup"
    assert optimized.imports[0].module == "pe"
    assert yara_file.includes[0].path == "common.yar"


def test_memory_optimizer_transformer_leaf_visitors_are_passthrough_or_pool() -> None:
    pool: dict[str, str] = {}
    transformer = MemoryOptimizerTransformer(pool, aggressive=False)

    str_lit = StringLiteral("hello")
    ident = Identifier("hello")
    string_id = StringIdentifier("$a")
    wildcard = StringWildcard("$a*")
    binary = BinaryExpression(BooleanLiteral(True), "and", IntegerLiteral(1))
    unary = UnaryExpression("not", Identifier("x"))
    plain = PlainString(identifier="$p", value="hello")
    regex = RegexString(identifier="$r", regex="hello")
    hexs = HexString(identifier="$h", tokens=[HexByte(0x41)])

    result_str = transformer.visit_string_literal(str_lit)
    result_ident = transformer.visit_identifier(ident)
    result_sid = transformer.visit_string_identifier(string_id)
    result_wc = transformer.visit_string_wildcard(wildcard)
    result_bin = transformer.visit_binary_expression(binary)
    result_un = transformer.visit_unary_expression(unary)
    transformer.visit_hex_string(hexs)
    result_regex = transformer.visit_regex_string(regex)
    transformer.visit_plain_string(plain)
    assert transformer.visit_boolean_literal(BooleanLiteral(True)).value is True
    assert transformer.visit_integer_literal(IntegerLiteral(1)).value == 1
    assert transformer.visit_double_literal(DoubleLiteral(1.5)).value == 1.5

    assert pool["hello"] == "hello"
    assert result_ident.name == "hello"
    assert result_str.value == "hello"
    assert result_sid.name == "$a"
    assert result_wc.pattern == "$a*"
    assert result_bin.operator == "and"
    assert result_un.operator == "not"
    assert result_regex.regex == "hello"
    assert plain.identifier == "$p"
    assert hexs.identifier == "$h"



"""Additional coverage for extern AST nodes."""

from __future__ import annotations

from yaraast.ast.extern import (
    ExternImport,
    ExternNamespace,
    ExternRule,
    ExternRuleReference,
    create_extern_rule,
)


class _Visitor:
    def visit_extern_rule(self, node):
        return ("rule", node.name)

    def visit_extern_rule_reference(self, node):
        return ("ref", node.qualified_name)

    def visit_extern_import(self, node):
        return ("import", node.module_path)

    def visit_extern_namespace(self, node):
        return ("namespace", node.name)


def test_extern_accepts_and_non_selective_import_str() -> None:
    visitor = _Visitor()

    rule = ExternRule(name="R1")
    ref = ExternRuleReference(rule_name="R1")
    imp = ExternImport(module_path="ext_rules")
    ns = ExternNamespace(name="ext")

    assert rule.accept(visitor) == ("rule", "R1")
    assert ref.accept(visitor) == ("ref", "R1")
    assert imp.accept(visitor) == ("import", "ext_rules")
    assert ns.accept(visitor) == ("namespace", "ext")
    assert str(imp) == 'import "ext_rules"'


def test_extern_namespace_negative_lookup_and_factory_without_modifiers() -> None:
    ns = ExternNamespace(name="ext")
    rule = ExternRule(name="R1")
    ns.add_extern_rule(rule)

    assert ns.get_rule_by_name("missing") is None
    assert str(ns) == "namespace ext"

    helper_rule = create_extern_rule("R2")
    assert helper_rule.modifiers == []
    assert helper_rule.namespace is None



"""Tests for formatting configuration.

Copyright (c) 2025 Marc Rivero López
Licensed under GPLv3. See LICENSE file for details.
This test suite validates real code behavior without mocks or stubs.
"""

from __future__ import annotations

from yaraast.codegen.formatting import (
    BraceStyle,
    FormattingConfig,
    HexStyle,
    IndentStyle,
    PredefinedStyles,
    StringStyle,
)


class TestFormattingEnums:
    """Test formatting enum definitions."""

    def test_indent_style_enum_values(self) -> None:
        """IndentStyle should have SPACES and TABS."""
        assert IndentStyle.SPACES.value == "spaces"
        assert IndentStyle.TABS.value == "tabs"

    def test_brace_style_enum_values(self) -> None:
        """BraceStyle should have correct values."""
        assert BraceStyle.SAME_LINE.value == "same_line"
        assert BraceStyle.NEW_LINE.value == "new_line"
        assert BraceStyle.K_AND_R.value == "k_and_r"

    def test_string_style_enum_values(self) -> None:
        """StringStyle should have correct values."""
        assert StringStyle.COMPACT.value == "compact"
        assert StringStyle.ALIGNED.value == "aligned"
        assert StringStyle.TABULAR.value == "tabular"

    def test_hex_style_enum_values(self) -> None:
        """HexStyle should have correct values."""
        assert HexStyle.LOWERCASE.value == "lowercase"
        assert HexStyle.UPPERCASE.value == "uppercase"


class TestFormattingConfigDefaults:
    """Test default configuration values."""

    def test_config_default_initialization(self) -> None:
        """FormattingConfig should initialize with sensible defaults."""
        config = FormattingConfig()

        assert config.indent_style == IndentStyle.SPACES
        assert config.indent_size == 4
        assert config.brace_style == BraceStyle.SAME_LINE
        assert config.space_before_colon is True
        assert config.space_after_colon is True
        assert config.space_around_operators is True
        assert config.space_after_comma is True
        assert config.string_style == StringStyle.ALIGNED
        assert config.align_string_modifiers is True
        assert config.hex_style == HexStyle.LOWERCASE
        assert config.hex_group_size == 0
        assert config.blank_lines_between_rules == 1
        assert config.blank_lines_between_sections == 1
        assert config.max_line_length == 120
        assert config.sort_imports is False
        assert config.sort_rules is False
        assert config.sort_meta is False
        assert config.sort_strings is False
        assert config.section_order == ["meta", "strings", "condition"]
        assert config.preserve_comments is True
        assert config.comment_style == "//"

    def test_config_custom_initialization(self) -> None:
        """FormattingConfig should accept custom values."""
        config = FormattingConfig(
            indent_style=IndentStyle.TABS,
            indent_size=2,
            brace_style=BraceStyle.NEW_LINE,
            space_before_colon=False,
        )

        assert config.indent_style == IndentStyle.TABS
        assert config.indent_size == 2
        assert config.brace_style == BraceStyle.NEW_LINE
        assert config.space_before_colon is False


class TestFormattingConfigSerialization:
    """Test configuration serialization."""

    def test_to_dict_includes_all_fields(self) -> None:
        """To_dict should include all configuration fields."""
        config = FormattingConfig()

        result = config.to_dict()

        assert "indent_style" in result
        assert "indent_size" in result
        assert "brace_style" in result
        assert "space_before_colon" in result
        assert "space_after_colon" in result
        assert "space_around_operators" in result
        assert "space_after_comma" in result
        assert "string_style" in result
        assert "align_string_modifiers" in result
        assert "hex_style" in result
        assert "hex_group_size" in result
        assert "blank_lines_between_rules" in result
        assert "blank_lines_between_sections" in result
        assert "max_line_length" in result
        assert "sort_imports" in result
        assert "sort_rules" in result
        assert "sort_meta" in result
        assert "sort_strings" in result
        assert "section_order" in result
        assert "preserve_comments" in result
        assert "comment_style" in result

    def test_to_dict_converts_enums_to_values(self) -> None:
        """To_dict should convert enum values to strings."""
        config = FormattingConfig(indent_style=IndentStyle.TABS, brace_style=BraceStyle.NEW_LINE)

        result = config.to_dict()

        assert result["indent_style"] == "tabs"
        assert result["brace_style"] == "new_line"
        assert isinstance(result["indent_style"], str)
        assert isinstance(result["brace_style"], str)

    def test_to_dict_preserves_primitive_types(self) -> None:
        """To_dict should preserve int, bool, and list types."""
        config = FormattingConfig(
            indent_size=2,
            space_before_colon=False,
            section_order=["condition", "strings", "meta"],
        )

        result = config.to_dict()

        assert result["indent_size"] == 2
        assert isinstance(result["indent_size"], int)
        assert result["space_before_colon"] is False
        assert isinstance(result["space_before_colon"], bool)
        assert result["section_order"] == ["condition", "strings", "meta"]
        assert isinstance(result["section_order"], list)

    def test_from_dict_creates_config(self) -> None:
        """From_dict should create FormattingConfig from dictionary."""
        data = {
            "indent_style": "tabs",
            "indent_size": 2,
            "brace_style": "new_line",
            "space_before_colon": False,
        }

        config = FormattingConfig.from_dict(data)

        assert isinstance(config, FormattingConfig)
        assert config.indent_style == IndentStyle.TABS
        assert config.indent_size == 2
        assert config.brace_style == BraceStyle.NEW_LINE
        assert config.space_before_colon is False

    def test_from_dict_with_all_fields(self) -> None:
        """From_dict should handle all configuration fields."""
        data = {
            "indent_style": "spaces",
            "indent_size": 4,
            "brace_style": "same_line",
            "space_before_colon": True,
            "space_after_colon": True,
            "space_around_operators": True,
            "space_after_comma": True,
            "string_style": "aligned",
            "align_string_modifiers": True,
            "hex_style": "uppercase",
            "hex_group_size": 8,
            "blank_lines_between_rules": 2,
            "blank_lines_between_sections": 1,
            "max_line_length": 100,
            "sort_imports": True,
            "sort_rules": True,
            "sort_meta": True,
            "sort_strings": True,
            "section_order": ["meta", "condition", "strings"],
            "preserve_comments": False,
            "comment_style": "/*",
        }

        config = FormattingConfig.from_dict(data)

        assert config.indent_style == IndentStyle.SPACES
        assert config.indent_size == 4
        assert config.hex_style == HexStyle.UPPERCASE
        assert config.hex_group_size == 8
        assert config.max_line_length == 100
        assert config.sort_imports is True
        assert config.section_order == ["meta", "condition", "strings"]
        assert config.preserve_comments is False
        assert config.comment_style == "/*"

    def test_from_dict_with_partial_data(self) -> None:
        """From_dict should use defaults for missing fields."""
        data = {"indent_size": 2, "brace_style": "new_line"}

        config = FormattingConfig.from_dict(data)

        assert config.indent_size == 2
        assert config.brace_style == BraceStyle.NEW_LINE
        # Should use defaults for missing fields
        assert config.indent_style == IndentStyle.SPACES
        assert config.space_before_colon is True

    def test_from_dict_with_empty_dict(self) -> None:
        """From_dict should work with empty dictionary."""
        config = FormattingConfig.from_dict({})

        assert isinstance(config, FormattingConfig)
        assert config.indent_style == IndentStyle.SPACES
        assert config.indent_size == 4

    def test_roundtrip_serialization(self) -> None:
        """Config should survive to_dict/from_dict roundtrip."""
        original = FormattingConfig(
            indent_style=IndentStyle.TABS,
            indent_size=2,
            brace_style=BraceStyle.K_AND_R,
            hex_style=HexStyle.UPPERCASE,
            section_order=["condition", "meta", "strings"],
        )

        dict_repr = original.to_dict()
        restored = FormattingConfig.from_dict(dict_repr)

        assert restored.indent_style == original.indent_style
        assert restored.indent_size == original.indent_size
        assert restored.brace_style == original.brace_style
        assert restored.hex_style == original.hex_style
        assert restored.section_order == original.section_order


class TestPredefinedStylesCompact:
    """Test compact predefined style."""

    def test_compact_style_minimal_spacing(self) -> None:
        """Compact style should have minimal spacing."""
        config = PredefinedStyles.compact()

        assert config.indent_size == 2
        assert config.brace_style == BraceStyle.SAME_LINE
        assert config.space_before_colon is False
        assert config.space_after_colon is False
        assert config.space_around_operators is False
        assert config.space_after_comma is False
        assert config.string_style == StringStyle.COMPACT
        assert config.blank_lines_between_rules == 0
        assert config.blank_lines_between_sections == 0

    def test_compact_via_class_method(self) -> None:
        """FormattingConfig.compact should call PredefinedStyles.compact."""
        config = FormattingConfig.compact()

        assert config.indent_size == 2
        assert config.brace_style == BraceStyle.SAME_LINE
        assert config.space_before_colon is False


class TestPredefinedStylesReadable:
    """Test readable predefined style."""

    def test_readable_style_balanced_spacing(self) -> None:
        """Readable style should have balanced spacing."""
        config = PredefinedStyles.readable()

        assert config.indent_size == 4
        assert config.brace_style == BraceStyle.SAME_LINE
        assert config.space_before_colon is True
        assert config.space_after_colon is True
        assert config.space_around_operators is True
        assert config.space_after_comma is True
        assert config.string_style == StringStyle.ALIGNED
        assert config.blank_lines_between_rules == 1
        assert config.blank_lines_between_sections == 1


class TestPredefinedStylesVerbose:
    """Test verbose predefined style."""

    def test_verbose_style_maximum_readability(self) -> None:
        """Verbose style should maximize readability."""
        config = PredefinedStyles.verbose()

        assert config.indent_size == 4
        assert config.brace_style == BraceStyle.NEW_LINE
        assert config.space_before_colon is True
        assert config.space_after_colon is True
        assert config.space_around_operators is True
        assert config.space_after_comma is True
        assert config.string_style == StringStyle.TABULAR
        assert config.align_string_modifiers is True
        assert config.blank_lines_between_rules == 2
        assert config.blank_lines_between_sections == 1
        assert config.sort_imports is True
        assert config.sort_meta is True
        assert config.sort_strings is True

    def test_expanded_via_class_method(self) -> None:
        """FormattingConfig.expanded should call PredefinedStyles.verbose."""
        config = FormattingConfig.expanded()

        assert config.indent_size == 4
        assert config.brace_style == BraceStyle.NEW_LINE
        assert config.sort_imports is True


class TestPredefinedStylesYaraDefault:
    """Test YARA default predefined style."""

    def test_yara_default_style(self) -> None:
        """YARA default style should match typical YARA formatting."""
        config = PredefinedStyles.yara_default()

        assert config.indent_size == 2
        assert config.brace_style == BraceStyle.SAME_LINE
        assert config.space_before_colon is True
        assert config.space_after_colon is True
        assert config.space_around_operators is True
        assert config.space_after_comma is True
        assert config.string_style == StringStyle.ALIGNED


class TestFormattingConfigIndentation:
    """Test indentation configuration."""

    def test_spaces_indentation(self) -> None:
        """Config should support space-based indentation."""
        config = FormattingConfig(indent_style=IndentStyle.SPACES, indent_size=4)

        assert config.indent_style == IndentStyle.SPACES
        assert config.indent_size == 4

    def test_tabs_indentation(self) -> None:
        """Config should support tab-based indentation."""
        config = FormattingConfig(indent_style=IndentStyle.TABS, indent_size=1)

        assert config.indent_style == IndentStyle.TABS
        assert config.indent_size == 1

    def test_custom_indent_size(self) -> None:
        """Config should support custom indent sizes."""
        config2 = FormattingConfig(indent_size=2)
        config8 = FormattingConfig(indent_size=8)

        assert config2.indent_size == 2
        assert config8.indent_size == 8


class TestFormattingConfigBraceStyles:
    """Test brace style configuration."""

    def test_same_line_brace_style(self) -> None:
        """Config should support same-line brace style."""
        config = FormattingConfig(brace_style=BraceStyle.SAME_LINE)

        assert config.brace_style == BraceStyle.SAME_LINE

    def test_new_line_brace_style(self) -> None:
        """Config should support new-line brace style."""
        config = FormattingConfig(brace_style=BraceStyle.NEW_LINE)

        assert config.brace_style == BraceStyle.NEW_LINE

    def test_k_and_r_brace_style(self) -> None:
        """Config should support K&R brace style."""
        config = FormattingConfig(brace_style=BraceStyle.K_AND_R)

        assert config.brace_style == BraceStyle.K_AND_R


class TestFormattingConfigSpacing:
    """Test spacing configuration."""

    def test_spacing_options_all_enabled(self) -> None:
        """Config should support all spacing options enabled."""
        config = FormattingConfig(
            space_before_colon=True,
            space_after_colon=True,
            space_around_operators=True,
            space_after_comma=True,
        )

        assert config.space_before_colon is True
        assert config.space_after_colon is True
        assert config.space_around_operators is True
        assert config.space_after_comma is True

    def test_spacing_options_all_disabled(self) -> None:
        """Config should support all spacing options disabled."""
        config = FormattingConfig(
            space_before_colon=False,
            space_after_colon=False,
            space_around_operators=False,
            space_after_comma=False,
        )

        assert config.space_before_colon is False
        assert config.space_after_colon is False
        assert config.space_around_operators is False
        assert config.space_after_comma is False


class TestFormattingConfigStringStyles:
    """Test string style configuration."""

    def test_compact_string_style(self) -> None:
        """Config should support compact string style."""
        config = FormattingConfig(string_style=StringStyle.COMPACT)

        assert config.string_style == StringStyle.COMPACT

    def test_aligned_string_style(self) -> None:
        """Config should support aligned string style."""
        config = FormattingConfig(string_style=StringStyle.ALIGNED)

        assert config.string_style == StringStyle.ALIGNED

    def test_tabular_string_style(self) -> None:
        """Config should support tabular string style."""
        config = FormattingConfig(string_style=StringStyle.TABULAR)

        assert config.string_style == StringStyle.TABULAR

    def test_align_string_modifiers(self) -> None:
        """Config should support modifier alignment toggle."""
        enabled = FormattingConfig(align_string_modifiers=True)
        disabled = FormattingConfig(align_string_modifiers=False)

        assert enabled.align_string_modifiers is True
        assert disabled.align_string_modifiers is False


class TestFormattingConfigHexStyles:
    """Test hex string style configuration."""

    def test_lowercase_hex_style(self) -> None:
        """Config should support lowercase hex style."""
        config = FormattingConfig(hex_style=HexStyle.LOWERCASE)

        assert config.hex_style == HexStyle.LOWERCASE

    def test_uppercase_hex_style(self) -> None:
        """Config should support uppercase hex style."""
        config = FormattingConfig(hex_style=HexStyle.UPPERCASE)

        assert config.hex_style == HexStyle.UPPERCASE

    def test_hex_group_size_options(self) -> None:
        """Config should support various hex group sizes."""
        no_grouping = FormattingConfig(hex_group_size=0)
        group_by_4 = FormattingConfig(hex_group_size=4)
        group_by_8 = FormattingConfig(hex_group_size=8)

        assert no_grouping.hex_group_size == 0
        assert group_by_4.hex_group_size == 4
        assert group_by_8.hex_group_size == 8


class TestFormattingConfigLineBreaks:
    """Test line break configuration."""

    def test_blank_lines_between_rules(self) -> None:
        """Config should support blank lines between rules."""
        none_config = FormattingConfig(blank_lines_between_rules=0)
        single_config = FormattingConfig(blank_lines_between_rules=1)
        double_config = FormattingConfig(blank_lines_between_rules=2)

        assert none_config.blank_lines_between_rules == 0
        assert single_config.blank_lines_between_rules == 1
        assert double_config.blank_lines_between_rules == 2

    def test_blank_lines_between_sections(self) -> None:
        """Config should support blank lines between sections."""
        config = FormattingConfig(blank_lines_between_sections=1)

        assert config.blank_lines_between_sections == 1

    def test_max_line_length(self) -> None:
        """Config should support max line length configuration."""
        config80 = FormattingConfig(max_line_length=80)
        config120 = FormattingConfig(max_line_length=120)

        assert config80.max_line_length == 80
        assert config120.max_line_length == 120


class TestFormattingConfigSorting:
    """Test sorting configuration."""

    def test_sort_options_all_disabled(self) -> None:
        """Config should support all sorting disabled."""
        config = FormattingConfig(
            sort_imports=False, sort_rules=False, sort_meta=False, sort_strings=False
        )

        assert config.sort_imports is False
        assert config.sort_rules is False
        assert config.sort_meta is False
        assert config.sort_strings is False

    def test_sort_options_all_enabled(self) -> None:
        """Config should support all sorting enabled."""
        config = FormattingConfig(
            sort_imports=True, sort_rules=True, sort_meta=True, sort_strings=True
        )

        assert config.sort_imports is True
        assert config.sort_rules is True
        assert config.sort_meta is True
        assert config.sort_strings is True

    def test_selective_sorting(self) -> None:
        """Config should support selective sorting."""
        config = FormattingConfig(sort_imports=True, sort_meta=True)

        assert config.sort_imports is True
        assert config.sort_meta is True
        assert config.sort_rules is False
        assert config.sort_strings is False


class TestFormattingConfigSectionOrder:
    """Test section order configuration."""

    def test_default_section_order(self) -> None:
        """Config should have default section order."""
        config = FormattingConfig()

        assert config.section_order == ["meta", "strings", "condition"]

    def test_custom_section_order(self) -> None:
        """Config should support custom section order."""
        custom_order = ["condition", "meta", "strings"]
        config = FormattingConfig(section_order=custom_order)

        assert config.section_order == custom_order

    def test_section_order_modification(self) -> None:
        """Section order should be modifiable."""
        config = FormattingConfig()
        config.section_order = ["strings", "condition", "meta"]

        assert config.section_order == ["strings", "condition", "meta"]


class TestFormattingConfigComments:
    """Test comment configuration."""

    def test_preserve_comments_enabled(self) -> None:
        """Config should support comment preservation."""
        config = FormattingConfig(preserve_comments=True)

        assert config.preserve_comments is True

    def test_preserve_comments_disabled(self) -> None:
        """Config should support disabling comment preservation."""
        config = FormattingConfig(preserve_comments=False)

        assert config.preserve_comments is False

    def test_single_line_comment_style(self) -> None:
        """Config should support single-line comment style."""
        config = FormattingConfig(comment_style="//")

        assert config.comment_style == "//"

    def test_multi_line_comment_style(self) -> None:
        """Config should support multi-line comment style."""
        config = FormattingConfig(comment_style="/*")

        assert config.comment_style == "/*"



"""CLI tests for LSP command (no mocks)."""

from __future__ import annotations

import pytest
from click.testing import CliRunner

from yaraast.cli.commands.lsp import lsp


def test_lsp_reports_missing_dependency() -> None:
    try:
        from yaraast.lsp import server as _server  # noqa: F401
    except Exception:
        runner = CliRunner()
        result = runner.invoke(lsp, [])
        assert result.exit_code != 0
        assert "Missing dependency" in result.output
        return

    pytest.skip("LSP dependencies available; skipping to avoid starting server.")



"""Label helpers for string pattern diagrams."""

from __future__ import annotations

from typing import Any


class StringDiagramLabelsMixin:
    """Mixin providing string diagram helpers."""

    def _truncate(self, text: str, max_len: int) -> str:
        if len(text) <= max_len:
            return text
        return text[:max_len] + "..."

    def _label_lines(self, *lines: str) -> str:
        return "\\n".join(lines)

    def _create_pattern_label(self, pattern_info: dict[str, Any]) -> str:
        """Create label for plain string pattern."""
        identifier = pattern_info["identifier"]
        value = pattern_info.get("value", "")
        length = pattern_info.get("length", 0)

        # Truncate long values
        display_value = self._truncate(value, 20)

        return self._label_lines(identifier, f'"{display_value}"', f"Length: {length}")

    def _create_hex_pattern_label(self, pattern_info: dict[str, Any]) -> str:
        """Create label for hex pattern."""
        identifier = pattern_info["identifier"]
        tokens = pattern_info.get("tokens", 0)
        token_analysis = pattern_info.get("token_analysis", {})

        wildcards = token_analysis.get("wildcards", 0)
        wildcard_ratio = token_analysis.get("wildcard_ratio", 0)

        return f"{identifier}\\nTokens: {tokens}\\nWildcards: {wildcards} ({wildcard_ratio:.1%})"

    def _create_regex_pattern_label(self, pattern_info: dict[str, Any]) -> str:
        """Create label for regex pattern."""
        identifier = pattern_info["identifier"]
        pattern = pattern_info.get("pattern", "")
        regex_analysis = pattern_info.get("regex_analysis", {})

        # Truncate long patterns
        display_pattern = self._truncate(pattern, 15)
        groups = regex_analysis.get("groups", 0)

        return self._label_lines(identifier, f"/{display_pattern}/", f"Groups: {groups}")

    def _create_short_label(self, pattern_info: dict[str, Any]) -> str:
        """Create short label for similarity diagram."""
        return pattern_info["identifier"]

    def _create_hex_token_label(self, token_analysis: dict[str, Any]) -> str:
        """Create detailed hex token label."""
        bytes_count = token_analysis.get("bytes", 0)
        wildcards = token_analysis.get("wildcards", 0)
        jumps = token_analysis.get("jumps", 0)
        alternatives = token_analysis.get("alternatives", 0)

        return f"Bytes: {bytes_count}|Wildcards: {wildcards}|Jumps: {jumps}|Alternatives: {alternatives}"

    def _create_hex_complexity_label(
        self,
        pattern_info: dict[str, Any],
        token_analysis: dict[str, Any],
    ) -> str:
        """Create hex complexity metrics label."""
        complexity = self._calculate_pattern_complexity(pattern_info)
        wildcard_ratio = token_analysis.get("wildcard_ratio", 0)
        complexity_score = token_analysis.get("complexity_score", 0)

        return self._label_lines(
            f"Overall Complexity: {complexity}",
            f"Wildcard Ratio: {wildcard_ratio:.1%}",
            f"Token Score: {complexity_score}",
        )

    def _get_pattern_shape(self, pattern_type: str) -> str:
        """Get shape for pattern type."""
        if pattern_type == "plain":
            return "circle"
        if pattern_type == "hex":
            return "hexagon"
        return "ellipse"



"""Module-related types and environment."""

from __future__ import annotations

from dataclasses import dataclass

from ._registry_base import YaraType
from .module_contracts import ModuleDefinition


@dataclass
class ModuleType(YaraType):
    """Module type with attributes."""

    module_name: str
    attributes: dict[str, YaraType]

    def __str__(self) -> str:
        return f"module({self.module_name})"

    def is_compatible_with(self, other: YaraType) -> bool:
        return isinstance(other, ModuleType) and other.module_name == self.module_name

    def get_attribute_type(self, attr: str) -> YaraType | None:
        """Get type of module attribute."""
        return self.attributes.get(attr)


@dataclass
class FunctionType(YaraType):
    """Function type."""

    name: str
    param_types: list[YaraType]
    return_type: YaraType

    def __str__(self) -> str:
        params = ", ".join(str(p) for p in self.param_types)
        return f"{self.name}({params}) -> {self.return_type}"

    def is_compatible_with(self, other: YaraType) -> bool:
        return False


class TypeSystem:
    """Type system with module support."""

    def __init__(self) -> None:
        self.modules: dict[str, ModuleDefinition] = {}
        self._init_modules()

    def _init_modules(self) -> None:
        """Initialize modules using ModuleLoader."""
        try:
            from yaraast.types.module_loader import ModuleLoader

            loader = ModuleLoader()
            self.modules = loader.modules
        except ImportError:
            self._init_builtin_modules()

    def _init_builtin_modules(self) -> None:
        """Initialize builtin modules (fallback)."""
        from .module_definitions import load_builtin_modules

        builtins = load_builtin_modules()
        self.modules = {name: builtins[name] for name in ("pe", "math") if name in builtins}

    def get_module(self, name: str) -> ModuleDefinition | None:
        """Get module definition by name."""
        return self.modules.get(name)



"""Workspace symbols provider for YARAAST LSP."""

from __future__ import annotations

import logging
from pathlib import Path

from lsprotocol.types import SymbolInformation

from yaraast.lsp.runtime import DocumentContext, LspRuntime, path_to_uri

logger = logging.getLogger(__name__)


class WorkspaceSymbolsProvider:
    """Provide workspace-wide symbol search."""

    def __init__(self, runtime: LspRuntime | None = None) -> None:
        """Initialize workspace symbols provider."""
        self.runtime = runtime
        self.symbol_cache: dict[str, tuple[float, list[SymbolInformation]]] = {}
        self.workspace_root: Path | None = None

    def set_workspace_root(self, root_path: str) -> None:
        """Set the workspace root directory."""
        self.workspace_root = Path(root_path)

    def get_workspace_symbols(self, query: str) -> list[SymbolInformation]:
        """Search for symbols across the entire workspace."""
        if self.runtime:
            return self.runtime.workspace_symbols(query)
        if not self.workspace_root or not self.workspace_root.exists():
            return []

        symbols = []

        # Find all YARA files in workspace
        yara_files = list(self.workspace_root.rglob("*.yar")) + list(
            self.workspace_root.rglob("*.yara")
        )

        for yara_file in yara_files:
            try:
                # Get symbols from this file
                file_symbols = self._get_symbols_from_file(yara_file)

                # Filter by query (case-insensitive substring match)
                if query:
                    query_lower = query.lower()
                    file_symbols = [sym for sym in file_symbols if query_lower in sym.name.lower()]

                symbols.extend(file_symbols)

            except Exception:
                logger.debug("Operation failed in %s", __name__, exc_info=True)
                # Skip files that fail to parse
                continue

        return symbols

    def _get_symbols_from_file(self, file_path: Path) -> list[SymbolInformation]:
        """Extract all symbols from a YARA file."""
        # Check cache first
        cache_key = str(file_path)
        mtime = file_path.stat().st_mtime

        if cache_key in self.symbol_cache:
            cached_mtime, cached_symbols = self.symbol_cache[cache_key]
            if cached_mtime == mtime:
                return cached_symbols

        # Parse file and extract symbols
        symbols = []

        try:
            with open(file_path) as f:
                content = f.read()

            file_uri = path_to_uri(file_path)
            doc = DocumentContext(file_uri, content)
            for record in doc.symbols():
                if record.kind not in {"rule", "string"}:
                    continue
                info = record.to_symbol_information()
                if record.kind == "rule":
                    info = SymbolInformation(
                        name=info.name,
                        kind=info.kind,
                        location=info.location,
                        container_name=file_path.name,
                    )
                elif record.kind == "string":
                    info = SymbolInformation(
                        name=info.name,
                        kind=info.kind,
                        location=info.location,
                        container_name=f"{file_path.name} :: {record.container_name}",
                    )
                symbols.append(info)

            # Cache results
            self.symbol_cache[cache_key] = (mtime, symbols)

        except Exception:
            logger.debug("Operation failed in %s", __name__, exc_info=True)
            # Return empty list if parsing fails
            pass

        return symbols

    def clear_cache(self) -> None:
        """Clear the symbol cache."""
        if self.runtime:
            return
        self.symbol_cache.clear()

    def invalidate_file(self, file_path: str) -> None:
        """Invalidate cache for a specific file."""
        if self.runtime:
            return
        if file_path in self.symbol_cache:
            del self.symbol_cache[file_path]



"""Decorator for safe LSP operation handling."""

from __future__ import annotations

from collections.abc import Callable
import functools
import logging
from typing import Any, overload

logger = logging.getLogger(__name__)


@overload
def lsp_safe_handler[F: Callable[..., Any]](func: F) -> F: ...


@overload
def lsp_safe_handler[F: Callable[..., Any]](*, default: Any) -> Callable[[F], F]: ...


def lsp_safe_handler[F: Callable[..., Any]](
    func: F | None = None, *, default: Any = None
) -> F | Callable[[F], F]:
    """Wrap LSP handler to catch exceptions and return a default value.

    LSP handlers use broad ``except Exception`` intentionally: the language
    server must remain responsive even when individual operations fail on
    malformed input.  Narrowing to domain exceptions would risk crashing
    the server on unexpected edge cases.

    Can be used as:
        @lsp_safe_handler          # returns None on error
        @lsp_safe_handler(default=[])  # returns [] on error
    """

    def decorator(fn: F) -> F:
        @functools.wraps(fn)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            try:
                return fn(*args, **kwargs)
            except Exception:
                logger.debug(
                    "Operation failed in %s.%s",
                    fn.__module__,
                    fn.__qualname__,
                    exc_info=True,
                )
                return default

        return wrapper  # type: ignore[return-value]

    if func is not None:
        return decorator(func)
    return decorator



"""Real tests for libyara availability paths (no mocks)."""

from __future__ import annotations

import pytest

from yaraast.libyara.compiler import YARA_AVAILABLE as COMPILER_AVAILABLE
from yaraast.libyara.compiler import LibyaraCompiler
from yaraast.libyara.scanner import YARA_AVAILABLE as SCANNER_AVAILABLE
from yaraast.libyara.scanner import LibyaraScanner


def test_libyara_compiler_import_error_when_missing() -> None:
    if COMPILER_AVAILABLE:
        pytest.skip("yara-python available; skipping ImportError check")
    with pytest.raises(ImportError):
        LibyaraCompiler()


def test_libyara_scanner_import_error_when_missing() -> None:
    if SCANNER_AVAILABLE:
        pytest.skip("yara-python available; skipping ImportError check")
    with pytest.raises(ImportError):
        LibyaraScanner()



"""Real tests for fluent condition builder helpers (no mocks)."""

from __future__ import annotations

from yaraast.builder.fluent_condition_builder import FluentConditionBuilder


def test_fluent_condition_quantifiers() -> None:
    builder = FluentConditionBuilder()
    expr = builder.any_of_them().build()
    assert "any" in str(expr)

    expr2 = builder.all_of_them().build()
    assert "all" in str(expr2)

    expr3 = builder.not_them().build()
    assert "not" in str(expr3)


def test_fluent_condition_string_helpers() -> None:
    builder = FluentConditionBuilder()
    expr = builder.string_count_eq("$a", 2).build()
    assert "==" in str(expr)

    expr2 = builder.string_at_entrypoint("$a").build()
    assert "entrypoint" in str(expr2)

    expr3 = builder.string_in_first_kb("$a").build()
    assert "1024" in str(expr3)



"""Additional tests for small helper modules without mocks."""

from __future__ import annotations

import struct

from yaraast.ast.modifiers import StringModifier
from yaraast.codegen.generator_helpers import (
    escape_plain_string_value,
    format_integer_literal,
    format_modifiers,
)
from yaraast.dialects import YaraDialect, detect_dialect
from yaraast.evaluation.evaluation_helpers import (
    BUILTIN_READERS,
    LITTLE_ENDIAN_ALIASES,
    _read_int8,
    _read_int16,
    _read_int16_be,
    _read_int32,
    _read_int32_be,
    _read_uint8,
    _read_uint16,
    _read_uint16_be,
    _read_uint32,
    _read_uint32_be,
    read_struct,
)
from yaraast.lexer.lexer import Lexer
from yaraast.lexer.lexer_helpers import (
    _skip_block_comment,
    _skip_line_comment,
    _skip_line_continuation,
    skip_whitespace_and_comments,
)


def test_lexer_helpers_skip_whitespace_comments_and_line_continuation() -> None:
    lex = Lexer("  \t\r\n// line comment\n/* block */rule")
    skip_whitespace_and_comments(lex)
    assert lex._current_char() == "r"

    combined = Lexer("\\  \n// c\n/* b */x")
    skip_whitespace_and_comments(combined)
    assert combined._current_char() == "x"

    cont = Lexer("\\  \r\nx")
    _skip_line_continuation(cont)
    assert cont._current_char() == "x"

    cont_cr = Lexer("\\\rx")
    _skip_line_continuation(cont_cr)
    assert cont_cr._current_char() == "x"

    cont_lf = Lexer("\\\nx")
    _skip_line_continuation(cont_lf)
    assert cont_lf._current_char() == "x"

    line = Lexer("// comment only\nx")
    _skip_line_comment(line)
    assert line._current_char() == "\n"

    block = Lexer("/* comment */x")
    _skip_block_comment(block)
    assert block._current_char() == "x"

    unterminated = Lexer("/* unterminated")
    _skip_block_comment(unterminated)
    assert unterminated.position == len(unterminated.text)  # Consumes entire unterminated comment


def test_evaluation_helpers_read_struct_and_builtin_readers() -> None:
    data = bytes([0x7F]) + struct.pack("<H", 0x1234) + struct.pack("<I", 0x12345678)
    data += struct.pack("b", -2) + struct.pack("<h", -1234) + struct.pack("<i", -56789)
    data += struct.pack(">H", 0xBEEF) + struct.pack(">I", 0xA1B2C3D4)
    data += struct.pack(">h", -2222) + struct.pack(">i", -333333)

    assert read_struct(data, "B", -1, 1) == 0
    assert read_struct(data, "I", len(data), 4) == 0

    assert _read_uint8(data, 0) == 0x7F
    assert _read_uint16(data, 1) == 0x1234
    assert _read_uint32(data, 3) == 0x12345678
    assert _read_int8(data, 7) == -2
    assert _read_int16(data, 8) == -1234
    assert _read_int32(data, 10) == -56789
    assert _read_uint16_be(data, 14) == 0xBEEF
    assert _read_uint32_be(data, 16) == 0xA1B2C3D4
    assert _read_int16_be(data, 20) == -2222
    assert _read_int32_be(data, 22) == -333333

    assert BUILTIN_READERS["uint16"](data, 1) == 0x1234
    assert BUILTIN_READERS[LITTLE_ENDIAN_ALIASES["uint16le"]](data, 1) == 0x1234
    assert LITTLE_ENDIAN_ALIASES["int32le"] == "int32"


def test_generator_helpers_escape_integer_and_modifiers() -> None:
    escaped = escape_plain_string_value('a\\b"c\n\r\t\x00\x01')
    assert escaped == 'a\\\\b\\"c\\n\\r\\t\\x00\\x01'

    assert format_integer_literal("not-a-number") == "not-a-number"
    assert format_integer_literal("1024") == "0x400"
    assert format_integer_literal(0x4D5A) == "0x4D5A"
    assert format_integer_literal(0x30) == "48"
    assert format_integer_literal(0x200) == "0x200"
    assert format_integer_literal(255) == "255"

    mod = StringModifier.from_name_value("ascii")
    assert format_modifiers(["wide", mod, 7], lambda node: f"<{node.name}>") == " wide <ascii> 7"
    assert format_modifiers(("nocase",), lambda node: "") == " nocase"
    assert format_modifiers("private", lambda node: "") == ""  # bare strings no longer accepted
    assert format_modifiers({"x"}, lambda node: "") == ""
    assert format_modifiers(123, lambda node: "") == ""


def test_detect_dialect_yaral_signals_and_default_yara() -> None:
    structural = """
rule login_event {
  events:
    $e.metadata.event_type = "LOGIN"
  condition:
    $e
}
"""
    assert detect_dialect(structural) == YaraDialect.YARA_L

    udm = 'rule x { condition: $e.principal.hostname == "host" }'
    assert detect_dialect(udm) == YaraDialect.YARA_L

    agg = 'rule x { condition: count_distinct("src") > 2 }'
    assert detect_dialect(agg) == YaraDialect.YARA_L

    plain_yara = 'rule classic { strings: $a = "events:" condition: $a }'
    assert detect_dialect(plain_yara) == YaraDialect.YARA


def test_detect_dialect_yarax_signals() -> None:
    yarax_lambda = "rule x { condition: lambda x, y: x + y }"
    assert detect_dialect(yarax_lambda) == YaraDialect.YARA_X

    yarax_match = "rule x { condition: match x { 1 => 2, _ => 3 } }"
    assert detect_dialect(yarax_match) == YaraDialect.YARA_X

    yarax_array_comp = "rule x { condition: [x for x in items if x] }"
    assert detect_dialect(yarax_array_comp) == YaraDialect.YARA_X

    yarax_with = 'rule x { condition: with $a = "test", $b = 2: true }'
    assert detect_dialect(yarax_with) == YaraDialect.YARA_X



"""Additional real coverage for ast_diff_compare helpers."""

from __future__ import annotations

from yaraast.ast.expressions import BooleanLiteral
from yaraast.ast.rules import Import, Include, Rule, Tag
from yaraast.ast.strings import PlainString
from yaraast.serialization.ast_diff import DiffNode, DiffResult, DiffType
from yaraast.serialization.ast_diff_compare import compare_imports, compare_includes, compare_rules
from yaraast.serialization.ast_diff_hasher import AstHasher


def test_compare_imports_and_includes_removed_and_alias_modified() -> None:
    result = DiffResult(old_ast_hash="old", new_ast_hash="new")

    old_imports = [Import(module="pe", alias="x"), Import(module="math")]
    new_imports = [Import(module="pe", alias="y")]
    compare_imports(old_imports, new_imports, result, DiffNode, DiffType)

    old_includes = [Include(path="old.yar"), Include(path="gone.yar")]
    new_includes = [Include(path="old.yar"), Include(path="new.yar")]
    compare_includes(old_includes, new_includes, result, DiffNode, DiffType)

    paths = {d.path: d for d in result.differences}
    assert paths["/imports/math"].diff_type == DiffType.REMOVED
    assert paths["/imports/pe/alias"].diff_type == DiffType.MODIFIED
    assert paths["/includes/gone.yar"].diff_type == DiffType.REMOVED
    assert paths["/includes/new.yar"].diff_type == DiffType.ADDED


def test_compare_rules_removed_modifier_and_string_diffs() -> None:
    hasher = AstHasher()
    result = DiffResult(old_ast_hash="old", new_ast_hash="new")

    old_rules = [
        Rule(
            name="old_only",
            condition=BooleanLiteral(value=True),
        ),
        Rule(
            name="shared",
            modifiers=["private"],
            strings=[
                PlainString(identifier="$a", value="x"),
                PlainString(identifier="$b", value="y"),
            ],
            condition=BooleanLiteral(value=True),
        ),
    ]
    new_rules = [
        Rule(
            name="shared",
            modifiers=["global"],
            strings=[PlainString(identifier="$a", value="x2")],
            condition=BooleanLiteral(value=True),
        ),
        Rule(
            name="new_only",
            condition=BooleanLiteral(value=True),
        ),
    ]

    compare_rules(old_rules, new_rules, result, hasher, DiffNode, DiffType)

    by_path = {d.path: d for d in result.differences}
    assert by_path["/rules/old_only"].diff_type == DiffType.REMOVED
    assert by_path["/rules/new_only"].diff_type == DiffType.ADDED
    assert by_path["/rules/shared/modifiers"].diff_type == DiffType.MODIFIED
    assert by_path["/rules/shared/strings/$b"].diff_type == DiffType.REMOVED
    assert by_path["/rules/shared/strings/$a"].diff_type == DiffType.MODIFIED


def test_compare_rules_meta_tags_and_added_string_diffs() -> None:
    hasher = AstHasher()
    result = DiffResult(old_ast_hash="old", new_ast_hash="new")

    old_rules = [
        Rule(
            name="shared",
            tags=[],
            meta={"author": "a"},
            strings=[],
            condition=BooleanLiteral(value=True),
        ),
    ]
    new_rules = [
        Rule(
            name="shared",
            tags=[Tag(name="tag1")],
            meta={"author": "b"},
            strings=[PlainString(identifier="$c", value="z")],
            condition=BooleanLiteral(value=True),
        ),
    ]

    compare_rules(old_rules, new_rules, result, hasher, DiffNode, DiffType)

    by_path = {d.path: d for d in result.differences}
    assert by_path["/rules/shared/meta"].diff_type == DiffType.MODIFIED
    assert by_path["/rules/shared/tags"].diff_type == DiffType.MODIFIED
    assert by_path["/rules/shared/strings/$c"].diff_type == DiffType.ADDED



"""Types and formatting helpers for the error-tolerant parser."""

from __future__ import annotations

from dataclasses import dataclass, field

from yaraast.ast.base import YaraFile


@dataclass
class ParserError:
    """Represents a parsing error."""

    message: str
    line: int
    column: int
    context: str = ""
    severity: str = "error"

    def format_error(self) -> str:
        """Format the error for display."""
        if self.context:
            return f"Line {self.line}:{self.column}: {self.message}\n  {self.context}"
        return f"Line {self.line}:{self.column}: {self.message}"


@dataclass
class ParseResult:
    """Result of error-tolerant parsing."""

    ast: YaraFile
    errors: list[ParserError] = field(default_factory=list)
    warnings: list[ParserError] = field(default_factory=list)


def format_parser_errors(errors: list[ParserError]) -> str:
    """Format a collection of parser errors for display."""
    if not errors:
        return "No errors"

    lines: list[str] = []
    for error in errors:
        lines.append(f"{error.severity.upper()}: {error.message}")
        lines.append(f"  at line {error.line}, column {error.column}")
        if error.context:
            lines.append(f"  > {error.context}")
            lines.append(f"    {' ' * error.column}^")
        lines.append("")
    return "\n".join(lines)



"""Analysis helpers for string pattern diagrams."""

from __future__ import annotations

from collections import defaultdict
import re
from typing import TYPE_CHECKING, Any

from yaraast.ast.strings import HexString, PlainString, RegexString
from yaraast.metrics.string_diagrams_common import modifier_names

if TYPE_CHECKING:
    from yaraast.ast.base import YaraFile


class StringDiagramAnalysisMixin:
    """Mixin providing string diagram helpers."""

    def _analyze_patterns(self, ast: YaraFile) -> None:
        """Analyze all string patterns in the AST."""
        self.string_patterns.clear()
        self.pattern_relationships.clear()
        self.pattern_stats.clear()

        pattern_id = 0
        for rule in ast.rules:
            self._current_rule = rule.name

            for string_def in rule.strings:
                pattern_id += 1
                pid = f"pattern_{pattern_id}"

                pattern_info: dict[str, Any] = {
                    "id": pid,
                    "identifier": string_def.identifier,
                    "rule": rule.name,
                    "modifiers": modifier_names(string_def.modifiers),
                }

                if isinstance(string_def, PlainString):
                    pattern_info.update(
                        {
                            "type": "plain",
                            "value": string_def.value,
                            "length": len(string_def.value),
                            "printable_ratio": self._calculate_printable_ratio(
                                string_def.value,
                            ),
                        },
                    )

                elif isinstance(string_def, HexString):
                    token_analysis = self._analyze_hex_tokens(string_def.tokens)
                    pattern_info.update(
                        {
                            "type": "hex",
                            "tokens": len(string_def.tokens),
                            "token_analysis": token_analysis,
                            "length": len(string_def.tokens),
                        },
                    )

                elif isinstance(string_def, RegexString):
                    regex_analysis = self._analyze_regex_pattern(string_def.regex)
                    pattern_info.update(
                        {
                            "type": "regex",
                            "pattern": string_def.regex,
                            "regex_analysis": regex_analysis,
                            "length": len(string_def.regex),
                        },
                    )

                self.string_patterns[pid] = pattern_info

    def _analyze_hex_tokens(self, tokens: list) -> dict[str, Any]:
        """Analyze hex string tokens."""
        from yaraast.ast.strings import HexAlternative, HexByte, HexJump, HexWildcard

        analysis = {
            "bytes": 0,
            "wildcards": 0,
            "jumps": 0,
            "alternatives": 0,
            "wildcard_ratio": 0.0,
            "complexity_score": 0,
        }

        for token in tokens:
            if isinstance(token, HexByte):
                analysis["bytes"] += 1
            elif isinstance(token, HexWildcard):
                analysis["wildcards"] += 1
            elif isinstance(token, HexJump):
                analysis["jumps"] += 1
                analysis["complexity_score"] += 2
            elif isinstance(token, HexAlternative):
                analysis["alternatives"] += 1
                analysis["complexity_score"] += 3

        total_tokens = len(tokens)
        if total_tokens > 0:
            analysis["wildcard_ratio"] = analysis["wildcards"] / total_tokens

        analysis["complexity_score"] += analysis["wildcards"] * 0.5

        return analysis

    def _analyze_regex_pattern(self, pattern: str) -> dict[str, Any]:
        """Analyze regex pattern complexity."""
        analysis = {
            "groups": len(re.findall(r"\([^?]", pattern)),
            "quantifiers": len(re.findall(r"[*+?{]", pattern)),
            "anchors": len(re.findall(r"[\^$]", pattern)),
            "character_classes": len(re.findall(r"\[[^\]]+\]", pattern)),
            "complexity_score": 0,
        }

        # Calculate complexity score
        analysis["complexity_score"] = (
            analysis["groups"] * 2
            + analysis["quantifiers"]
            + analysis["character_classes"] * 1.5
            + analysis["anchors"] * 0.5
        )

        return analysis

    def _calculate_printable_ratio(self, text: str) -> float:
        """Calculate ratio of printable characters."""
        if not text:
            return 0.0

        printable_count = sum(1 for c in text if c.isprintable() and not c.isspace())
        return printable_count / len(text)

    def _calculate_pattern_complexity(self, pattern_info: dict[str, Any]) -> int:
        """Calculate overall pattern complexity score."""
        complexity = 1  # Base complexity

        if pattern_info["type"] == "plain":
            # Plain string complexity based on length and printability
            length = pattern_info.get("length", 0)
            printable_ratio = pattern_info.get("printable_ratio", 1.0)

            complexity += length // 10  # Length factor
            if printable_ratio < 0.8:
                complexity += 2  # Non-printable penalty

        elif pattern_info["type"] == "hex":
            # Hex complexity from token analysis
            token_analysis = pattern_info.get("token_analysis", {})
            complexity += int(token_analysis.get("complexity_score", 0))

        elif pattern_info["type"] == "regex":
            # Regex complexity from analysis
            regex_analysis = pattern_info.get("regex_analysis", {})
            complexity += int(regex_analysis.get("complexity_score", 0))

        # Modifier complexity
        complexity += len(pattern_info.get("modifiers", []))

        return complexity

    def _find_similar_patterns(self) -> dict[str, set[str]]:
        """Find groups of similar patterns."""
        groups = defaultdict(set)

        # Group by type first
        for pattern_id, pattern_info in self.string_patterns.items():
            pattern_type = pattern_info["type"]
            groups[pattern_type].add(pattern_id)

        # Further group by characteristics
        refined_groups = {}

        for group_type, pattern_ids in groups.items():
            if group_type == "plain":
                # Group plain strings by length ranges
                short_patterns = set()
                medium_patterns = set()
                long_patterns = set()

                for pid in pattern_ids:
                    length = self.string_patterns[pid].get("length", 0)
                    if length < 10:
                        short_patterns.add(pid)
                    elif length < 50:
                        medium_patterns.add(pid)
                    else:
                        long_patterns.add(pid)

                if short_patterns:
                    refined_groups["Short Plain"] = short_patterns
                if medium_patterns:
                    refined_groups["Medium Plain"] = medium_patterns
                if long_patterns:
                    refined_groups["Long Plain"] = long_patterns

            elif group_type == "hex":
                # Group hex patterns by wildcard ratio
                low_wildcard = set()
                high_wildcard = set()

                for pid in pattern_ids:
                    token_analysis = self.string_patterns[pid].get("token_analysis", {})
                    wildcard_ratio = token_analysis.get("wildcard_ratio", 0)

                    if wildcard_ratio < 0.3:
                        low_wildcard.add(pid)
                    else:
                        high_wildcard.add(pid)

                if low_wildcard:
                    refined_groups["Precise Hex"] = low_wildcard
                if high_wildcard:
                    refined_groups["Flexible Hex"] = high_wildcard

            else:  # regex
                refined_groups["Regex"] = pattern_ids

        return refined_groups

    def _calculate_similarity(
        self,
        pattern1: dict[str, Any],
        pattern2: dict[str, Any],
    ) -> float:
        """Calculate similarity between two patterns."""
        if pattern1["type"] != pattern2["type"]:
            return 0.0

        similarity = 0.0

        if pattern1["type"] == "plain":
            # Simple string similarity
            str1 = pattern1.get("value", "")
            str2 = pattern2.get("value", "")

            if len(str1) == 0 or len(str2) == 0:
                return 0.0

            # Simple Jaccard similarity on character level
            set1 = set(str1.lower())
            set2 = set(str2.lower())
            intersection = len(set1 & set2)
            union = len(set1 | set2)

            similarity = intersection / union if union > 0 else 0.0

        elif pattern1["type"] == "hex":
            # Compare hex token patterns
            tokens1 = pattern1.get("token_analysis", {})
            tokens2 = pattern2.get("token_analysis", {})

            # Compare wildcard ratios
            ratio1 = tokens1.get("wildcard_ratio", 0)
            ratio2 = tokens2.get("wildcard_ratio", 0)
            similarity = 1.0 - abs(ratio1 - ratio2)

        elif pattern1["type"] == "regex":
            # Simple regex similarity (could be enhanced)
            regex1 = pattern1.get("pattern", "")
            regex2 = pattern2.get("pattern", "")

            if len(regex1) == 0 or len(regex2) == 0:
                return 0.0

            # Character-level similarity
            set1 = set(regex1)
            set2 = set(regex2)
            intersection = len(set1 & set2)
            union = len(set1 | set2)

            similarity = intersection / union if union > 0 else 0.0

        return similarity



"""AST-focused report generation for complexity analysis."""

from __future__ import annotations

from typing import Any

from yaraast.ast.base import YaraFile
from yaraast.metrics.complexity import ComplexityAnalyzer
from yaraast.metrics.complexity_helpers import (
    calculate_cognitive_complexity,
    calculate_rule_complexity,
)


def generate_complexity_report(ast: YaraFile) -> dict[str, Any]:
    """Generate a comprehensive complexity report for a YARA file."""
    analyzer = ComplexityAnalyzer()
    metrics = analyzer.analyze(ast)

    rules_data = []
    for rule in ast.rules:
        complexity = calculate_rule_complexity(rule)
        cyclomatic = metrics.cyclomatic_complexity.get(rule.name, 1)
        cognitive = calculate_cognitive_complexity(rule.condition) if rule.condition else 0

        rules_data.append(
            {
                "name": rule.name,
                "total_complexity": complexity,
                "cyclomatic_complexity": cyclomatic,
                "cognitive_complexity": cognitive,
                "strings": len(rule.strings) if rule.strings else 0,
                "modifiers": len(rule.modifiers),
            },
        )

    return {
        "rules": rules_data,
        "summary": {
            "total_rules": len(ast.rules),
            "avg_complexity": sum(r["total_complexity"] for r in rules_data)
            / max(1, len(rules_data)),
            "max_complexity": max(
                (r["total_complexity"] for r in rules_data),
                default=0,
            ),
            "quality_score": metrics.get_quality_score(),
            "quality_grade": metrics.get_complexity_grade(),
        },
        "metrics": metrics.to_dict(),
    }



"""Reporting helpers for string diffs."""

from __future__ import annotations


def emit_string_added(
    base_path: str,
    result,
    diff_node,
    diff_type,
    identifier: str,
) -> None:
    """Record added string."""
    result.differences.append(
        diff_node(
            path=f"{base_path}/{identifier}",
            diff_type=diff_type.ADDED,
            new_value=identifier,
            node_type="StringDefinition",
        ),
    )


def emit_string_removed(
    base_path: str,
    result,
    diff_node,
    diff_type,
    identifier: str,
) -> None:
    """Record removed string."""
    result.differences.append(
        diff_node(
            path=f"{base_path}/{identifier}",
            diff_type=diff_type.REMOVED,
            old_value=identifier,
            node_type="StringDefinition",
        ),
    )


def emit_string_modified(
    base_path: str,
    result,
    diff_node,
    diff_type,
    identifier: str,
    old_hash: str,
    new_hash: str,
) -> None:
    """Record modified string."""
    result.differences.append(
        diff_node(
            path=f"{base_path}/{identifier}",
            diff_type=diff_type.MODIFIED,
            old_value=old_hash,
            new_value=new_hash,
            node_type="StringDefinition",
        ),
    )



"""Extra real coverage for complexity helpers and model."""

from __future__ import annotations

from yaraast.ast.conditions import ForExpression, ForOfExpression
from yaraast.ast.expressions import BinaryExpression, Identifier, IntegerLiteral, UnaryExpression
from yaraast.ast.rules import Rule
from yaraast.ast.strings import HexByte, HexString, PlainString, RegexString
from yaraast.metrics.complexity_helpers import (
    calculate_cognitive_complexity,
    calculate_cyclomatic_complexity,
    calculate_expression_complexity,
    calculate_rule_complexity,
)
from yaraast.metrics.complexity_model import ComplexityMetrics


def test_complexity_helpers_cover_rule_expression_and_cognitive_paths() -> None:
    expr = BinaryExpression(
        left=UnaryExpression(operator="not", operand=Identifier(name="x")),
        operator="and",
        right=IntegerLiteral(value=1),
    )

    rule = Rule(
        name="helper_rule",
        modifiers=["private", "global"],
        strings=[
            PlainString(identifier="$a", value="abc"),
            HexString(identifier="$b", tokens=[HexByte(value=0x41)]),
            RegexString(identifier="$c", regex="ab+"),
        ],
        condition=expr,
    )

    assert calculate_rule_complexity(rule) > 1
    assert calculate_expression_complexity(expr) >= 1
    assert calculate_cyclomatic_complexity(expr) >= 2

    for_expr = ForExpression(
        quantifier="any",
        variable="i",
        iterable=IntegerLiteral(value=1),
        body=Identifier(name="i"),
    )
    for_of_expr = ForOfExpression(
        quantifier="all",
        string_set=Identifier(name="them"),
        condition=IntegerLiteral(value=1),
    )
    assert calculate_cognitive_complexity(for_expr) >= 4
    assert calculate_cognitive_complexity(for_of_expr) >= 4


def test_complexity_model_to_dict_quality_score_and_grades() -> None:
    metrics = ComplexityMetrics(
        total_rules=10,
        total_imports=1,
        total_includes=2,
        rules_with_strings=8,
        rules_with_meta=9,
        rules_with_tags=3,
        private_rules=1,
        global_rules=1,
        total_strings=12,
        plain_strings=6,
        hex_strings=3,
        regex_strings=3,
        strings_with_modifiers=4,
        max_condition_depth=9,
        avg_condition_depth=3.5,
        total_binary_ops=7,
        total_unary_ops=2,
        for_expressions=1,
        for_of_expressions=1,
        of_expressions=1,
        hex_wildcards=2,
        hex_jumps=1,
        hex_alternatives=1,
        regex_groups=3,
        regex_quantifiers=5,
        unused_strings=["r:$a", "r:$b"],
        complex_rules=["r1", "r2"],
        cyclomatic_complexity={"r1": 11},
        string_dependencies={"r1": {"$a", "$b"}},
        module_usage={"pe": 1},
    )

    data = metrics.to_dict()
    assert data["file_metrics"]["total_rules"] == 10
    assert sorted(data["dependencies"]["string_dependencies"]["r1"]) == ["$a", "$b"]
    assert data["quality_metrics"]["complex_rules"] == ["r1", "r2"]

    score = metrics.get_quality_score()
    assert score == 55.0
    assert metrics.get_complexity_grade() == "F"

    grade_cases = [
        (ComplexityMetrics(total_rules=1, rules_with_meta=1), "A"),
        (
            ComplexityMetrics(
                total_rules=10,
                rules_with_meta=0,
                max_condition_depth=6,
                unused_strings=["a"],
            ),
            "B",
        ),
        (
            ComplexityMetrics(
                total_rules=10,
                rules_with_meta=0,
                max_condition_depth=9,
                unused_strings=["a"],
            ),
            "C",
        ),
        (
            ComplexityMetrics(
                total_rules=10,
                rules_with_meta=0,
                max_condition_depth=6,
                unused_strings=["a", "b", "c", "d"],
                complex_rules=["r1"],
            ),
            "D",
        ),
        (
            ComplexityMetrics(
                total_rules=10,
                rules_with_meta=0,
                max_condition_depth=9,
                unused_strings=["a", "b", "c", "d"],
                complex_rules=["r1", "r2"],
            ),
            "F",
        ),
    ]

    for case_metrics, expected in grade_cases:
        assert case_metrics.get_complexity_grade() == expected



"""Branch-focused tests for AST transformer helpers (no mocks)."""

from __future__ import annotations

from yaraast.ast.base import YaraFile
from yaraast.ast.expressions import BinaryExpression, Identifier, StringIdentifier
from yaraast.ast.rules import Import, Include, Rule, Tag
from yaraast.ast.strings import PlainString
from yaraast.builder.ast_transformer import (
    CloneTransformer,
    RuleTransformer,
    YaraFileTransformer,
    clone_rule,
    clone_yara_file,
    create_rule_collection,
    create_variant_rule,
    merge_yara_files,
    transform_rule,
    transform_yara_file,
)


def _sample_rule(name: str = "r1") -> Rule:
    return Rule(
        name=name,
        modifiers=["global"],
        tags=[Tag(name="t1")],
        meta={"author": "me"},
        strings=[PlainString(identifier="$a", value="x")],
        condition=StringIdentifier(name="$a"),
    )


def test_clone_helpers_create_deep_independent_copies() -> None:
    original_rule = _sample_rule("orig")
    cloned_rule = clone_rule(original_rule)

    cloned_rule.name = "changed"
    cloned_rule.tags[0].name = "changed_tag"
    cloned_rule.meta[0].value = "other"
    cloned_rule.strings[0].identifier = "$b"

    assert original_rule.name == "orig"
    assert original_rule.tags[0].name == "t1"
    assert original_rule.get_meta_value("author") == "me"
    assert original_rule.strings[0].identifier == "$a"

    original_file = YaraFile(
        imports=[Import(module="pe")],
        includes=[Include(path="inc.yar")],
        rules=[original_rule],
    )
    cloned_file = clone_yara_file(original_file)
    cloned_file.imports[0].module = "math"
    cloned_file.rules[0].name = "mutated"

    assert original_file.imports[0].module == "pe"
    assert original_file.rules[0].name == "orig"


def test_rule_transformer_tag_modifier_meta_and_string_helpers() -> None:
    transformed = (
        RuleTransformer(_sample_rule())
        .rename("base")
        .add_prefix("pre_")
        .add_suffix("_suf")
        .add_tag("t2")
        .replace_tag("t1", "new")
        .remove_tag("missing")
        .add_modifier("private")
        .add_modifier("private")
        .remove_modifier("global")
        .set_author("you")
        .set_description("desc")
        .set_version(7)
        .remove_meta("missing")
        .rename_strings({"$a": "$renamed"})
        .build()
    )

    assert transformed.name == "pre_base_suf"
    assert {t.name for t in transformed.tags} == {"new", "t2"}
    assert [str(m) for m in transformed.modifiers] == ["private"]
    assert transformed.get_meta_value("author") == "you"
    assert transformed.get_meta_value("description") == "desc"
    assert transformed.get_meta_value("version") == 7
    assert transformed.strings[0].identifier == "$renamed"
    assert isinstance(transformed.condition, StringIdentifier)
    assert transformed.condition.name == "$renamed"


def test_rule_transformer_prefix_suffix_and_condition_transform_paths() -> None:
    rule = Rule(
        name="ids",
        strings=[
            PlainString(identifier="$one", value="1"),
            PlainString(identifier="two", value="2"),
        ],
        condition=BinaryExpression(
            left=Identifier(name="filesize"),
            operator=">",
            right=Identifier(name="entrypoint"),
        ),
    )
    transformer = RuleTransformer(rule)

    prefixed = transformer.prefix_strings("p_").suffix_strings("_s").build()
    assert [s.identifier for s in prefixed.strings] == ["$p_one_s", "p_two_s"]
    assert isinstance(prefixed.condition, BinaryExpression)

    with_added = (
        RuleTransformer(prefixed).add_string(PlainString(identifier="$extra", value="z")).build()
    )
    with_removed = RuleTransformer(with_added).remove_string("$extra").build()
    assert [s.identifier for s in with_removed.strings] == ["$p_one_s", "p_two_s"]

    cond_transformed = (
        RuleTransformer(with_removed)
        .transform_condition(
            lambda expr: BinaryExpression(left=expr, operator="and", right=Identifier(name="ok"))
        )
        .build()
    )
    assert isinstance(cond_transformed.condition, BinaryExpression)
    assert cond_transformed.condition.operator == "and"

    replaced = (
        RuleTransformer(cond_transformed)
        .replace_condition(StringIdentifier(name="$p_one_s"))
        .build()
    )
    assert isinstance(replaced.condition, StringIdentifier)


def test_yara_file_transformer_operations_and_filters() -> None:
    original = YaraFile(
        imports=[Import(module="pe")],
        includes=[Include(path="a.yar")],
        rules=[_sample_rule("one"), _sample_rule("two")],
    )

    transformed = (
        YaraFileTransformer(original)
        .add_import("pe")
        .add_import("math")
        .remove_import("missing")
        .add_include("a.yar")
        .add_include("b.yar")
        .remove_include("missing")
        .remove_rule("missing")
        .transform_rule("one", lambda r: RuleTransformer(r).rename("renamed").build())
        .prefix_all_rules("pre_")
        .suffix_all_rules("_suf")
        .add_tag_to_all_rules("all")
        .make_all_rules_private()
        .set_author_for_all_rules("team")
        .filter_by_tag("all")
        .filter_by_modifier("private")
        .build()
    )

    assert {imp.module for imp in transformed.imports} == {"pe", "math"}
    assert {inc.path for inc in transformed.includes} == {"a.yar", "b.yar"}
    assert [r.name for r in transformed.rules] == ["pre_renamed_suf", "pre_two_suf"]
    assert all(any(str(m) == "private" for m in r.modifiers) for r in transformed.rules)
    assert all(r.get_meta_value("author") == "team" for r in transformed.rules)


def test_convenience_transform_functions_and_variant_collection_merge_paths() -> None:
    base = _sample_rule("base")

    via_helper = transform_rule(base).make_public().make_global().build()
    assert not any(str(m) == "private" for m in via_helper.modifiers)
    assert any(str(m) == "global" for m in via_helper.modifiers)

    variant = create_variant_rule(
        base,
        "variant",
        prefix="pre_",
        suffix="_suf",
        tags=["x", "y"],
        author="author2",
        description="desc2",
        private=True,
    )
    assert variant.name == "pre_variant_suf"
    assert {t.name for t in variant.tags} == {"t1", "x", "y"}
    assert variant.get_meta_value("author") == "author2"
    assert variant.get_meta_value("description") == "desc2"
    assert any(str(m) == "private" for m in variant.modifiers)

    collection = create_rule_collection([base, variant], "grp")
    assert [r.name for r in collection.rules] == ["grp_base", "grp_pre_variant_suf"]

    empty_merge = merge_yara_files()
    assert empty_merge.rules == []
    assert empty_merge.imports == []

    f1 = YaraFile(imports=[Import(module="pe")], includes=[Include(path="x.yar")], rules=[base])
    f2 = YaraFile(
        imports=[Import(module="pe"), Import(module="math")],
        includes=[Include(path="x.yar"), Include(path="y.yar")],
        rules=[variant],
    )
    merged = merge_yara_files(f1, f2)
    assert {imp.module for imp in merged.imports} == {"pe", "math"}
    assert {inc.path for inc in merged.includes} == {"x.yar", "y.yar"}
    assert [r.name for r in merged.rules] == ["base", "pre_variant_suf"]

    helper_transformed = transform_yara_file(merged).remove_include("x.yar").build()
    assert {inc.path for inc in helper_transformed.includes} == {"y.yar"}

    direct_clone = CloneTransformer.clone_rule(base)
    deep_clone = CloneTransformer.clone(base)
    assert isinstance(direct_clone, Rule)
    assert isinstance(deep_clone, Rule)



"""Visitor helpers for MemoryOptimizerTransformer."""

from __future__ import annotations

import copy
from typing import TYPE_CHECKING

from yaraast.performance.memory_helpers import pooled_value

if TYPE_CHECKING:
    from yaraast.ast.base import ASTNode, YaraFile
    from yaraast.ast.expressions import (
        BinaryExpression,
        Identifier,
        StringIdentifier,
        StringLiteral,
        StringWildcard,
        UnaryExpression,
    )
    from yaraast.ast.meta import Meta
    from yaraast.ast.rules import Import, Include, Rule, Tag
    from yaraast.ast.strings import HexString, PlainString, RegexString


def _shallow(node: ASTNode) -> ASTNode:
    """Create a shallow copy of a dataclass node to avoid mutating the original."""
    return copy.copy(node)


def visit_string_literal(transformer, node: StringLiteral) -> StringLiteral:
    node = _shallow(node)
    if hasattr(node, "value") and isinstance(node.value, str):
        node.value = pooled_value(transformer.string_pool, node.value)
    return node


def visit_identifier(transformer, node: Identifier) -> Identifier:
    node = _shallow(node)
    if hasattr(node, "name") and isinstance(node.name, str):
        node.name = pooled_value(transformer.string_pool, node.name)
    return node


def visit_rule(transformer, node: Rule) -> Rule:
    node = _shallow(node)
    if node.name:
        node.name = pooled_value(transformer.string_pool, node.name)
    if node.condition:
        node.condition = transformer.visit(node.condition)
    if node.strings:
        node.strings = [transformer.visit(s) for s in node.strings]
    if node.meta:
        node.meta = [
            transformer.visit(m) if hasattr(m, "accept") else transformer.visit_meta(m)
            for m in node.meta
        ]
    if node.tags:
        node.tags = [transformer.visit(t) for t in node.tags]
    if transformer.aggressive and hasattr(node, "location"):
        node.location = None
    return node


def visit_plain_string(transformer, node: PlainString) -> PlainString:
    node = _shallow(node)
    if hasattr(node, "value") and isinstance(node.value, str):
        node.value = pooled_value(transformer.string_pool, node.value)
    if hasattr(node, "identifier") and isinstance(node.identifier, str):
        node.identifier = pooled_value(transformer.string_pool, node.identifier)
    return node


def visit_meta(transformer, node: Meta) -> Meta:
    node = _shallow(node)
    if hasattr(node, "key") and isinstance(node.key, str):
        node.key = pooled_value(transformer.string_pool, node.key)
    if hasattr(node, "value") and isinstance(node.value, str):
        node.value = pooled_value(transformer.string_pool, node.value)
    return node


def visit_tag(transformer, node: Tag) -> Tag:
    node = _shallow(node)
    if hasattr(node, "name") and isinstance(node.name, str):
        node.name = pooled_value(transformer.string_pool, node.name)
    return node


def visit_yara_file(transformer, node: YaraFile) -> YaraFile:
    node = _shallow(node)
    if node.imports:
        node.imports = [transformer.visit(imp) for imp in node.imports]
    if node.includes:
        node.includes = [transformer.visit(inc) for inc in node.includes]
    if node.rules:
        node.rules = [transformer.visit(rule) for rule in node.rules]
    return node


def visit_import(transformer, node: Import) -> Import:
    node = _shallow(node)
    if hasattr(node, "module") and isinstance(node.module, str):
        node.module = pooled_value(transformer.string_pool, node.module)
    return node


def visit_include(transformer, node: Include) -> Include:
    node = _shallow(node)
    if hasattr(node, "path") and isinstance(node.path, str):
        node.path = pooled_value(transformer.string_pool, node.path)
    return node


def visit_string_identifier(transformer, node: StringIdentifier) -> StringIdentifier:
    node = _shallow(node)
    if hasattr(node, "name") and isinstance(node.name, str):
        node.name = pooled_value(transformer.string_pool, node.name)
    return node


def visit_string_wildcard(transformer, node: StringWildcard) -> StringWildcard:
    node = _shallow(node)
    if hasattr(node, "pattern") and isinstance(node.pattern, str):
        node.pattern = pooled_value(transformer.string_pool, node.pattern)
    return node


def visit_binary_expression(transformer, node: BinaryExpression) -> BinaryExpression:
    node = _shallow(node)
    if hasattr(node, "left"):
        node.left = transformer.visit(node.left)
    if hasattr(node, "right"):
        node.right = transformer.visit(node.right)
    if hasattr(node, "operator") and isinstance(node.operator, str):
        node.operator = pooled_value(transformer.string_pool, node.operator)
    return node


def visit_unary_expression(transformer, node: UnaryExpression) -> UnaryExpression:
    node = _shallow(node)
    if hasattr(node, "operand"):
        node.operand = transformer.visit(node.operand)
    if hasattr(node, "operator") and isinstance(node.operator, str):
        node.operator = pooled_value(transformer.string_pool, node.operator)
    return node


def visit_hex_string(transformer, node: HexString) -> HexString:
    node = _shallow(node)
    if hasattr(node, "identifier") and isinstance(node.identifier, str):
        node.identifier = pooled_value(transformer.string_pool, node.identifier)
    return node


def visit_regex_string(transformer, node: RegexString) -> RegexString:
    node = _shallow(node)
    if hasattr(node, "identifier") and isinstance(node.identifier, str):
        node.identifier = pooled_value(transformer.string_pool, node.identifier)
    if hasattr(node, "regex") and isinstance(node.regex, str):
        node.regex = pooled_value(transformer.string_pool, node.regex)
    return node
